from __future__ import print_function

from burlap.common import Satchel
from burlap.tests.base import TestCase

class ManifestTests(TestCase):

    def test_get_current(self):

        class MyTestSatchel(Satchel):

            name = 'mytest'

            def set_defaults(self):
                self.env.myvar = 123

            def configure(self):
                pass

        mytest = MyTestSatchel()

        manifest = mytest.current_manifest
        print('manifest0:', manifest)
        assert manifest == {'myvar': 123, 'enabled': True}

        mytest.env.myvar = 456

        manifest = mytest.current_manifest
        print('manifest1:', manifest)
        assert manifest == {'myvar': 456, 'enabled': True}




from mock import patch

from burlap.tests.base import TestCase

class TestVagrantMachines(TestCase):

    def test_machines_one(self):
        with patch('burlap.vagrant.vagrant._status') as mock_status:
            mock_status.return_value = [('default', 'running')]
            from burlap.vagrant import vagrant
            self.assertEqual(vagrant.machines(), ['default'])



from __future__ import print_function

import os
import re
import sys
import datetime
import tempfile
import json
import functools
import traceback
import shutil
from collections import defaultdict
from pprint import pprint

import yaml

from fabric.api import env, sudo as _sudo, get as _get
import fabric.contrib.files
import fabric.api

from burlap import common
from burlap.common import (
    local_or_dryrun,
    put_or_dryrun,
    sudo_or_dryrun,
)
from burlap.decorators import task_or_dryrun, runs_once
from burlap import exceptions

LOCALHOSTS = ('localhost', '127.0.0.1')

STORAGE_LOCAL = 'local'
STORAGE_REMOTE = 'remote'
STORAGES = (
    STORAGE_LOCAL,
    STORAGE_REMOTE,
)

default_remote_path = '/var/local/burlap'

def init_env():
    env.plan_init = True
    env.plan_root = None
    env.plan_originals = {}
    env.plan_storage = STORAGE_REMOTE
    env.plan_lockfile_path = '/var/lock/burlap_deploy.lock'
    env.plan = None
    env.plan_data_dir = '%(burlap_data_dir)s/plans'
    env.plan_digits = 3

# Prevent globals from being reset by duplicate imports.
if not 'plan_init' in env:
    init_env()
    
_originals = env.plan_originals

RUN = 'run'
SUDO = 'sudo'
LOCAL = 'local'
PUT = 'put'
PLAN_METHODS = [RUN, SUDO, LOCAL, PUT]

INITIAL = 'initial'

_fs_cache = defaultdict(dict) # {func_name:{path:ret}}

def clear_fs_cache():
    _fs_cache.clear()
    RemoteFile._file_cache.clear()

def make_dir(d):
    if d not in _fs_cache['make_dir']:
        if env.plan_storage == STORAGE_REMOTE and env.host_string not in LOCALHOSTS:
            sudo_or_dryrun('mkdir -p "%s"' % d)
        else:
            if not os.path.isdir(d):
                os.makedirs(d)
        _fs_cache['make_dir'][d] = True
    return _fs_cache['make_dir'][d]

@task_or_dryrun
def list_dir(d):
    #print('list_dir.existing:', d, _fs_cache['list_dir'].get(d))
    if d not in _fs_cache['list_dir']:
        verbose = common.get_verbose()
        if env.plan_storage == STORAGE_REMOTE and env.host_string not in LOCALHOSTS:
            #print('listing remote path:', d)
            #output = sudo_or_dryrun('ls "%s"' % d)
            output = _sudo('ls "%s"' % d)
            output = output.split()
            if verbose:
                print('output:', output)
            ret = output
        else:
            #print('listing local path:', d)
            ret = os.listdir(d)
        _fs_cache['list_dir'][d] = ret
    return _fs_cache['list_dir'][d]

@task_or_dryrun
def is_dir(d):
    if d not in _fs_cache['is_dir']:
        verbose = common.get_verbose()
        if env.plan_storage == STORAGE_REMOTE and env.host_string not in LOCALHOSTS:
            cmd = 'if [ -d "%s" ]; then echo 1; else echo 0; fi' % d
            output = _sudo(cmd)
            if verbose:
                print('output:', output)
            #ret = int(output)
            ret = int(re.findall(r'^[0-9]+$', output, flags=re.DOTALL|re.I|re.M)[0])
        else:
            ret = os.path.isdir(d)
        _fs_cache['is_dir'][d] = ret
    return _fs_cache['is_dir'][d]

@task_or_dryrun
def is_file(fqfn):
    if fqfn not in _fs_cache['is_file']:
        verbose = common.get_verbose()
        if env.plan_storage == STORAGE_REMOTE and env.host_string not in LOCALHOSTS:
            cmd = 'if [ -f "%s" ]; then echo 1; else echo 0; fi' % fqfn
            output = _sudo(cmd)
            if verbose:
                print('output:', output)
            ret = int(re.findall(r'^[0-9]+$', output, flags=re.DOTALL|re.I|re.M)[0])
        else:
            ret = os.path.isfile(fqfn)
        _fs_cache['is_file'][fqfn] = ret
    return _fs_cache['is_file'][fqfn]

# class Singleton(type):
#     def __init__(cls, name, bases, dict):
#         print('singleton.init')
#         super(Singleton, cls).__init__(name, bases, dict)
#         cls.instance = None
#  
#     def __call__(cls, *args, **kw):
#         print('singleton.call')
#         return super(Singleton, cls).__call__(*args, **kw)
#         if cls.instance is None:
#             cls.instance = super(Singleton, cls).__call__(*args, **kw)
#         return cls.instance

class RemoteFile(object):
    """
    A helper class for allowing a remote file to be read and written locally
    while still ultimately being saved remotely.
    """
    
#     __metaclass__ = Singleton
    
    _file_cache = {} # {fqfn, obj}
    
    #TODO:use meta-class instead?
    #http://stackoverflow.com/questions/31875/is-there-a-simple-elegant-way-to-define-singletons-in-python/33201#33201
    
    def __new__(cls, fqfn, *args, **kwargs):
        # Remember and cache every class instance per unique file name.
        if fqfn not in cls._file_cache:
#             print('creating new instance:', fqfn)
            cls._file_cache[fqfn] = super(RemoteFile, cls).__new__(cls, fqfn, *args, **kwargs)
#         else:
#             print('using cache:', fqfn)
        return cls._file_cache[fqfn]
    
    def __init__(self, fqfn, mode='r'):
        super(RemoteFile, self).__init__() # causes instantiation error?
        
        assert mode in ('r', 'w', 'a'), 'Invalid mode: %s' % mode
        
        self.mode = mode
            
        # Due to the singleton-nature of __new__, this may be called multiple times,
        # so we check for and avoid duplicate calls.
        if not hasattr(self, 'fqfn'):
            
            self.fqfn = fqfn
            self.content = ''
            self.fresh = True
            
            if mode in 'ra':
    
                _, tmp_fn = tempfile.mkstemp()
                os.remove(tmp_fn)
                ret = _get(remote_path=fqfn, local_path=tmp_fn, use_sudo=True)
                #ret = get_or_dryrun(remote_path=fqfn, local_path=tmp_fn, use_sudo=True)
#                 print('ret:', ret)
                _fn = ret[0]
#                 print('reading:', _fn)
                fin = open(_fn, 'rb')
#                 print('reading2:', _fn)
                self.content = fin.read()
#                 print('closing')
                fin.close()
                #print('removing:', tmp_fn)
                os.remove(tmp_fn)#TODO:memory leak?
#                 print('done init load')
                
            if mode in 'wa':
                
                # Update file system cache.
                _fs_cache['is_file'][fqfn] = True
                
#         print('done init all')

    def write(self, s):
        assert self.mode in 'wa', 'File must be in write-mode.'
        self.content += s
        self.fresh = False
        # Note, flush() must to be called to actually write this.

    def read(self, *args, **kwargs):
        return self.content
    
    def readlines(self):
        return self.content.split('\n')
    
    def flush(self):
        if self.fresh:
            return
        
        print('Flushing contents to remote file.')
        _, tmp_fn = tempfile.mkstemp()
        os.remove(tmp_fn)
        fout = open(tmp_fn, 'w')
        #print('content:')
        #print(self.content)
        fout.write(self.content)
        fout.close()
        put_or_dryrun(
            local_path=tmp_fn,
            remote_path=self.fqfn,
            use_sudo=True)
        os.remove(tmp_fn)#TODO:memory leak?
        self.fresh = True
        
        # Update file system cache.
        _fs_cache['is_file'][self.fqfn] = True
        
    def close(self):
        print('Closing remote file.')
        self.flush()

def open_file(fqfn, mode='r'):
    verbose = common.get_verbose()
    if env.plan_storage == STORAGE_REMOTE and env.host_string not in LOCALHOSTS:
        return RemoteFile(fqfn, mode)
    else:
        return open(fqfn, mode)

def get_plan_data_dir():
    common.init_burlap_data_dir()
    d = env.plan_data_dir % env
    return d

def init_plan_data_dir():
    d = get_plan_data_dir()
    make_dir(d)
    return d

def delete_plan_data_dir():
    d = get_plan_data_dir()
    if env.plan_storage == STORAGE_REMOTE and env.host_string:
        _sudo('rm -Rf %s' % d)
    elif env.plan_storage == STORAGE_LOCAL:
        shutil.rmtree(d)

class Colors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

def fail(s):
    return Colors.FAIL + str(s) + Colors.ENDC

def success(s):
    return Colors.OKGREEN + str(s) + Colors.ENDC

def ongoing(s):
    return Colors.WARNING + str(s) + Colors.ENDC

def iter_plan_names(role=None):
    d = get_plan_dir(role=role)
    #print('iter_plan_names.d:', d)
    try:
        assert is_dir(d), 'Plan directory %s does not exist.' % d
    except AssertionError as e:
        #print('iter_plan_names.e:', e)
        if common.get_dryrun():
            # During dryrun, and the directory is missing, assume the host has been reset
            # and there are no prior plan files.
            return
        else:
            raise
    plan_subfolders = sorted(list_dir(d))
    #print('iter_plan_names.subfolders:', plan_subfolders)
    for name in plan_subfolders:
        #print('iter_plan_names.name:', iter_plan_names)
        fqfn = os.path.join(d, name)
        if not is_dir(fqfn):
            continue
        yield name

def get_thumbprint_path(role, name):
    d = get_plan_dir(role=role, name=name)
    d = os.path.join(d, 'thumbprints')
    make_dir(d)
    return d

def get_thumbprint_filename():
    return 'thumbprint'
        
class Step(object):
    """
    A single piece of a plan.
    """
    
    def __init__(self, command, host, method, user=None, key=None, args=None, kwargs=None):
        
        args = args or []
        kargs = kwargs or {}
        
        self.command = command
        self.host = host
        self.user = user
        self.method = method
        
        assert method in PLAN_METHODS
        
        self.key = key
        # The value entity attributes are set to as a result
        # of executing this step.
        self.args = args
        self.kwargs = kwargs
    
    @classmethod
    def from_line(cls, line):
        matches = re.findall(
            r'\[(?P<user>[^@]+)@(?P<host>[^\]]+)(?P<extra>{[^}]+})?]\s+(?P<method>[^\:]+):\s+(?P<command>.*?)$',
            line, flags=re.I|re.DOTALL)
        assert matches
        #print(matches)
        kwargs = dict(zip(['user', 'host', 'extra', 'method', 'command'], matches[0]))
        
        extra = {}
        if kwargs['extra']:
            extra = json.loads(kwargs['extra'])
        del kwargs['extra']
        
        if 'key' in extra:
            key = extra['key']
            del extra['key']
            kwargs['key'] = key
           
        step = Step(**kwargs)
        return step
    
    def execute(self):
        method = getattr(common, '%s_or_dryrun' % self.method)
        env.user = self.user
        env.host_string = self.host
        if self.key:
            env.key_filename = self.key
        else:
            env.key_filename = None
        method(self.command)
    
    def __str__(self):
        user_str = '%s@' % (self.user) if self.user else ''
        return '[%s%s] %s %s' % (
            user_str,
            self.host,
            self.method+':',
            self.command,
        )
    
    def __repr__(self):
        return str(self.__dict__)

HISTORY_HEADERS = ['step', 'start', 'end']

def get_plan_dir(role, name=None):
    role = role or env.ROLE
    assert role, 'No role defined.'
    if name:
        d = os.path.join(init_plan_data_dir(), role, name)
    else:
        d = os.path.join(init_plan_data_dir(), role)
    make_dir(d)
    return d

class Plan(object):
    """
    A sequence of steps for accomplishing a state change.
    """
    
    def __init__(self, name, role=None):
        
        self.verbose = verbose = common.get_verbose()
        
        self.name = name
        
        self.role = role or env.ROLE
        
        self.vprint('init plan dir')
        self.plan_dir = get_plan_dir(role, name)
        make_dir(self.plan_dir)
        assert is_dir(self.plan_dir)
        
        self.vprint('init plan history dir')
        self.plan_dir_history = os.path.join(self.plan_dir, 'history')
        if not is_file(self.plan_dir_history):
            fout = open_file(self.plan_dir_history, 'w')
            fout.write(','.join(HISTORY_HEADERS))
            fout.close()
        self.vprint('loading plan history')
        
        self.load_history()
        
        self.vprint('init plan index')
        self.plan_dir_index = os.path.join(self.plan_dir, 'index')
        if not is_file(self.plan_dir_index):
            fout = open_file(self.plan_dir_index, 'w')
            fout.write(str(0))
            fout.close()
        self.vprint('loading plan index')
        self.load_index()
        
        self.vprint('init plan steps')
        self.plan_dir_steps = os.path.join(self.plan_dir, 'steps')
        if not is_file(self.plan_dir_steps):
            fout = open_file(self.plan_dir_steps, 'w')
            fout.write('')
            fout.close()
        self.vprint('loading plan steps')
        self.load_steps()
        
        self.vprint('init plan hosts')
        self.plan_dir_hosts = os.path.join(self.plan_dir, 'hosts')
        if self.role == env.ROLE and not is_file(self.plan_dir_hosts):
            fout = open_file(self.plan_dir_hosts, 'w')
            fout.write('\n'.join(sorted(env.hosts)))
            fout.close()
        self.vprint('loading plan hosts')
        self.load_hosts()
        
        #self.plan_thumbprint_fn = os.path.join(self.plan_dir, 'thumbprint')
        
        self.vprint('plan init done')

    def vprint(self, *args, **kwargs):
        """
        When verbose is set, acts like the normal print() function.
        Otherwise, does nothing.
        """
        if common.get_verbose():
            print(*args, **kwargs)

    def __cmp__(self, other):
        if not isinstance(other, Plan):
            return NotImplemented
        return cmp((self.name, self.role), (other.name, other.role))
    
    def __unicode__(self):
        return unicode(self.name)
    
    def __repr__(self):
        return u'<%s: %s>' % (type(self).__name__, unicode(self))
    
    def is_complete(self):
        #return self.percent_complete == 100
        return self.all_hosts_thumbprinted and self.index == len(self._steps)
    
    @property
    def number(self):
        try:
            return int(re.findall('^[0-9]+', self.name)[0])
        except IndexError:
            #print('No number in "%s"' % self.name
            return 0
    
    def failed(self):
        return False #TODO
    
    def load_hosts(self):
        self.hosts = []
        if self.verbose: print('loading hosts, opening')
        fin = open_file(self.plan_dir_hosts, 'r')
        if self.verbose: print('loading hosts, readlines')
        lines = fin.readlines()
        if self.verbose: print('loading hosts, lines:', lines)
        for line in lines:
            if not line.strip():
                continue
            self.hosts.append(line.strip())
        if self.verbose: print('loading hosts, done')
    
    @property
    def all_hosts_thumbprinted(self):
        for host in self.hosts:
            fn = self.get_thumbprint_filename(host)
            if not is_file(fn):
                return False
        return True
    
    def load_history(self):
        pass
    
    def get_thumbprint_filename(self, host_string):
        d = os.path.join(self.plan_dir, 'thumbprints')
        make_dir(d)
        fn = os.path.join(d, env.host_string)
        return fn
    
    @property
    def thumbprint(self):
        verbose = common.get_verbose()
        self.vprint('plan.thumbprint')
        fn = self.get_thumbprint_filename(env.host_string)
        self.vprint('plan.thumbprint.fn:', fn)
        content = open_file(fn).read()
        self.vprint('plan.thumbprint.yaml.raw:', content)
        data = yaml.load(content)
        self.vprint('plan.thumbprint.yaml.data:', data)
        return data
    
    @thumbprint.setter
    def thumbprint(self, data):
        assert isinstance(data, dict)
        if not common.get_dryrun():
            fout = open_file(self.get_thumbprint_filename(env.host_string), 'w')
            yaml.dump(data, fout, default_flow_style=False, indent=4)
            fout.flush()
    
    def record_thumbprint(self, only_components=None):
        """
        Creates a thumbprint file for the current host in the current role and name.
        """
        #print('record_thumbprint.only_components:', only_components)
        only_components = only_components or []
        data = get_current_thumbprint(role=self.role, name=self.name, only_components=only_components)
        print('Recording thumbprint for host %s with deployment %s on %s.' % (env.host_string, self.name, self.role))
        #print('data:', data)
        self.thumbprint = data
    
    @property
    def remaining_step_count(self):
        return len(self._steps) - self.index
    
    def add_history(self, index, start, end):
        fout = open_file(self.plan_dir_history, 'a')
        fout.write('%s,%s,%s\n' % (index, start, end))
        fout.flush()
        fout.close()
    
    def load_index(self):
        self._index = int(open_file(self.plan_dir_index).read().strip())
    
    @property
    def index(self):
        return self._index
    
    def is_initial(self):
        return set(self.name) == set(['0'])
    
    @property
    def percent_complete(self):
        if self.is_initial():
            return 100
        if not self._steps:
            return 100
        return self.index/float(len(self._steps))*100
    
    @index.setter
    def index(self, v):
        self._index = int(v)
        fout = open_file(self.plan_dir_index, 'w')
        fout.write(str(self._index))
        fout.flush()
        fout.close()
    
    def load_steps(self):
        self._steps = []
        lines = open_file(self.plan_dir_steps).readlines()
        for line in lines:
            line = line.strip()
            if not line:
                continue
            elif '] executing task ' in line.lower():
                continue
            elif line.lower().startswith('done'):
                continue
            s = Step.from_line(line)
            self.add_step(s)

    @classmethod
    def get_or_create_next(cls, role=None, last_plan=None):
        role = role or env.ROLE
        last_plan = last_plan or get_last_plan(role=role)
        if last_plan:
            number = last_plan.number + 1
        else:
            number = 0
        assert len(str(number)) <= env.plan_digits, \
            'Too many deployments. Truncate existing or increase `plan_digits`.'
        plan = Plan(role=role, name=('%0'+str(env.plan_digits)+'i') % number)
        return plan

    @classmethod
    def load(cls, name, role=None):
        verbose = common.get_verbose()
        if verbose:
            print('loading plan:', name)
        plan = cls(name, role=role)
        return plan
    
    def add_step(self, s):
        assert isinstance(s, Step)
        self._steps.append(s)
    
    @property
    def steps(self):
        return list(self._steps)
    
    def clear(self):
        self.index = 0
        self._steps = []
    
    def execute(self, i=None, j=None):
        i = i or self.index
        steps_ran = []
        for step_i, step in enumerate(self.steps):
            if step_i < i:
                continue
                
            # Run command.
            t0 = datetime.datetime.utcnow().isoformat()
            step.execute()
            t1 = datetime.datetime.utcnow().isoformat()
            steps_ran.append(step)
            
            # Record success.
            if not common.get_dryrun():
                self.index = step_i + 1
                self.add_history(self.index, t0, t1)
            
            if j is not None and step_i >= j:
                break
                
        return steps_ran

#DEPRECATED
@task_or_dryrun
def record(name):
    common.set_dryrun(1)
    env.plan = Plan(name=name)

#DEPRECATED
@task_or_dryrun
def execute(name):
    verbose = common.get_verbose()
    plan = Plan.load(name, verbose=int(verbose))
    if verbose:
        if plan.is_complete():
            print('Execution of plan %s is complete.' % (plan.name,), file=sys.stderr)
        else:
            print('Execution of plan %s is %.02f%% complete.' % (plan.name, plan.percent_complete), file=sys.stderr)
    steps = []
    if not plan.is_complete():
        steps = plan.execute()
        if verbose:
            if plan.is_complete():
                print('Execution of plan %s is complete.' % (plan.name,), file=sys.stderr)
            else:
                print('Execution of plan %s is %.02f%% complete.' % (plan.name, plan.percent_complete), file=sys.stderr)
    if verbose:
        print('Executed %i steps.' % (len(steps),), file=sys.stderr)

def get_last_completed_plan():
    """
    Returns the last plan completed.
    """
    verbose = common.get_verbose()
    if verbose:
        print('get_last_completed_plan')
    for _name in reversed(sorted(list(iter_plan_names()))):
        #print('get_last_completed_plan._name:', _name)
        plan = Plan.load(_name)
        if verbose:
            print('plan:', plan.name)
            print('plan.completed:', plan.is_complete())
        if plan.is_complete():
            return plan
            
def get_last_plan(role=None):
    """
    Returns the last plan created.
    """
    for _name in reversed(sorted(list(iter_plan_names(role=role)))):
        plan = Plan.load(_name, role=role)
        return plan

@task_or_dryrun
def has_outstanding_plans():
    """
    Returns true if there are plans for this role that have not been executed.
    """
    verbose = common.get_verbose()
    last_completed = get_last_completed_plan()
    if verbose:
        print('last_completed plan:', last_completed)
    last = get_last_plan()
    if verbose:
        print('last plan:', last)
        print('eq:', last == last_completed)
    return last != last_completed

@task_or_dryrun
@runs_once
def status(name=None):
    """
    Reports the status of any pending plans for the current role.
    """
    print('plan,complete,percent')
    for _name in iter_plan_names():
        #print(_name)
        plan = Plan.load(_name)
        output = '%s,%s,%s' % (_name, int(plan.is_complete()), plan.percent_complete)
        if plan.is_complete():
            output = success(output)
        elif plan.failed:
            output = fail(output)
        else:
            output = ongoing(output)
        print(output)

def get_current_thumbprint(role=None, name=None, reraise=0, only_components=None):
    """
    Retrieves a snapshot of the current code state.
    """
    if name == INITIAL:
        name = '0'*env.plan_digits
    
    last = get_last_thumbprint()
    only_components = only_components or []
    only_components = [_.upper() for _ in only_components]
    data = {} # {component:data}
    manifest_data = (last and last.copy()) or {}
    for component_name, func in sorted(common.manifest_recorder.iteritems()):
        component_name = component_name.upper()
        #print('component_name:', component_name)
        
        if only_components and component_name not in only_components:
            if common.get_verbose():
                print('Skipping ignored component:', component_name)
            continue
            
        if component_name.lower() not in env.services:
            if common.get_verbose():
                print('Skipping unused component:', component_name)
            continue
            
        try:
            manifest_data[component_name] = func()
        except exceptions.AbortDeployment as e:
            raise
        except Exception as e:
            if int(reraise):
                raise
            print(traceback.format_exc(), file=sys.stderr)
        
    return manifest_data

@task_or_dryrun
def get_last_thumbprint():
    """
    Returns thumbprint from the last complete deployment.
    """
    verbose = common.get_verbose()
    plan = get_last_completed_plan()
    #print('get_last_thumbprint.last_completed_plan:', plan)
    if verbose: print('get_last_thumbprint.last completed plan:', plan and plan.name)
    last_thumbprint = (plan and plan.thumbprint) or {}
    if verbose: print('get_last_thumbprint.last_thumbprint:', last_thumbprint)
    return last_thumbprint

def iter_thumbprint_differences(only_components=None, local_verbose=0):
    only_components = only_components or []
    local_verbose = int(local_verbose)
    verbose = common.get_verbose() or local_verbose
    #if verbose: print('getting last thumbprint')
    last = get_last_thumbprint()
    #if verbose: print('getting current thumbprint')
    current = get_current_thumbprint()
    #if verbose: print('comparing thumbprints')
    for k in current:
        if only_components and k not in only_components:
#             print('iter_thumbprint_differences.skipping:', k)
            continue
#         print('iter:',k); raw_input('enter')
#         if verbose: print('iter_thumbprint_differences.NOT skipping:', k)
        if current[k] != last.get(k):
            if verbose:
                print('DIFFERENCE! k:', k, current[k], last.get(k))
                print('Current:')
                pprint(current[k], indent=4)
                print('Last:')
                pprint(last.get(k), indent=4)
            yield k, (last, current)
#     if verbose: print('iter_thumbprint_differences done')

@task_or_dryrun
def explain(name, **kwargs):
    #common.set_verbose(1)
    kwargs = kwargs or {}
    name = common.assert_valid_satchel(name)
    kwargs['only_components'] = [name]
    kwargs['local_verbose'] = 1
    diffs = dict(iter_thumbprint_differences(**kwargs))
    last, current = diffs.get(name, (None, None))
    if last is None and current is None:
        print('There are no differences.')
#     else:
#         last = last or {}
#         last.setdefault(name, {})
#         print('last:')
#         pprint(last[name], indent=4)
#         print('current:')
#         pprint(current[name], indent=4)

@task_or_dryrun
def show_diff(only=None):
    """
    Inspects differences between the last deployment and the current code state.
    """
    for k, (last, current) in iter_thumbprint_differences():
        if only and k.lower() != only.lower():
            continue
        print('Component %s has changed.' % k)
        last = last.get(k)
        current = current.get(k)
        if isinstance(last, dict) and isinstance(current, dict):
            for _k in set(last).union(current):
                _a = last.get(_k)
                _b = current.get(_k)
                if _a != _b:
                    print('DIFF: %s =' % _k, _a, _b)
        else:
            print('DIFF:', last, current)

@task_or_dryrun
def info():
    d = os.path.join(init_plan_data_dir(), env.ROLE)
    print('storage:', env.plan_storage)
    print('dir:', d)

@task_or_dryrun
def reset():
    """
    Deletes all recorded plan executions.
    This will cause the planner to think everything needs to be re-deployed.
    """
    d = os.path.join(init_plan_data_dir(), env.ROLE)
    if env.plan_storage == STORAGE_REMOTE and env.host_string not in LOCALHOSTS:
        sudo_or_dryrun('rm -Rf "%s"' % d)
        sudo_or_dryrun('mkdir -p "%s"' % d)
    elif env.plan_storage == STORAGE_LOCAL:
        local_or_dryrun('rm -Rf "%s"' % d)
        local_or_dryrun('mkdir -p "%s"' % d)
    else:
        raise NotImplementedError
    
@task_or_dryrun
@runs_once
def truncate():
    """
    Compacts all deployment records into a single initial deployment.
    """
    reset()
    if not common.get_dryrun():
        fabric.api.execute(thumbprint, hosts=env.hosts)

@task_or_dryrun
def thumbprint(name=None, components=None):
    """
    Creates a manifest file for the current host, listing all current settings
    so that a future deployment can use it as a reference to perform an
    idempotent deployment.
    """
    
    only_components = components or []
    if isinstance(only_components, basestring):
        only_components = [_.strip() for _ in only_components.split(',') if _.strip()]
    
    if name:
        plan = Plan.load(name=name)
    else:
        plan = get_last_plan()
        print('last plan:', plan)
        if not plan:
            plan = Plan.get_or_create_next()
    plan.record_thumbprint(only_components=only_components)
    
@task_or_dryrun
@runs_once
def preview(**kwargs):
    """
    Lists the likely pending deployment steps.
    """
    try:
        return auto(preview=1, **kwargs)
    except exceptions.AbortDeployment as e:
        common.print_fail('\n%s\n' % str(e))

def get_last_current_diffs(target_component):
    """
    Retrieves differing manifests between the current and last snapshot.
    """
    target_component = target_component.strip().upper()
    
    all_services = set(_.strip().upper() for _ in env.services)
    diffs = list(iter_thumbprint_differences())
    components = set()
    component_thumbprints = {}
    for component, (last, current) in diffs:
        if component not in all_services:
            continue
        component_thumbprints[component] = last, current
    
    print('component_thumbprints:', component_thumbprints.keys())
    last, current = component_thumbprints[target_component]
    return last, current

@task_or_dryrun
def auto(fake=0, preview=0, check_outstanding=1, components=None, explain=0, enable_plans=True, force=False):
    """
    Generates a plan based on the components that have changed since the last deployment.
    
    The overall steps ran for each host:
    
        1. create plan
        2. run plan
        3. create thumbprint
    
    fake := If true, generates the plan and records the run as successful, but does not apply any
        changes to the hosts.
    
    components := list of names of components found in the services list
    
    force := If true and specific components are specified, treats them as changed, even if no changes have been found
    """
    verbose = common.get_verbose()
    
    explain = int(explain)
    only_components = components or []
    if isinstance(only_components, basestring):
        only_components = [_.strip().upper() for _ in only_components.split(',') if _.strip()]
    if only_components:
        print('Limiting deployment to components: %s' % only_components)
    
    def get_deploy_funcs(components):
        for component in components:
            
            if only_components and component not in only_components:
                continue
            
            funcs = common.manifest_deployers.get(component, [])
            for func_name in funcs:
                
                #TODO:remove this after burlap.* naming prefix bug fixed
                if func_name.startswith('burlap.'):
                    print('skipping %s' % func_name)
                    continue
                    
                takes_diff = common.manifest_deployers_takes_diff.get(func_name, False)
#                 print(func_name, takes_diff)
                
                if preview:
                    #print(success((' '*4)+func_name))
                    #continue
                    yield func_name, None
                else:
                    func = common.resolve_deployer(func_name)
                    last, current = component_thumbprints[component]
                    if not fake:
                        if takes_diff:
                            yield func_name, functools.partial(func, last=last, current=current)
                        else:
                            yield func_name, functools.partial(func)

    fake = int(fake)
    preview = int(preview)
    check_outstanding = int(check_outstanding)
    
    all_services = set(_.strip().upper() for _ in env.services)
    if verbose:
        print('&'*80)
        print('services:', env.services)
    
    if enable_plans:
        last_plan = get_last_completed_plan()
        outstanding = has_outstanding_plans()
        if verbose:
            print('outstanding plans:', outstanding)
        if check_outstanding and outstanding:
            print(fail((
                'There are outstanding plans pending execution! '
                'Run `fab %s deploy.status` for details.\n'
                'To ignore these, re-run with :check_outstanding=0.'
            ) % env.ROLE))
            sys.exit(1)
    
    if verbose:
        print('iter_thumbprint_differences')
    diffs = list(iter_thumbprint_differences(only_components=only_components))
    if diffs:
        if verbose:
            print('Differences detected!')

    # Create plan.
    components = set()
    component_thumbprints = {}
    for component, (last, current) in diffs:
        if component not in all_services:
            print('ignoring component:', component)
            continue
#         if only_components and component not in only_components:
#             continue
        component_thumbprints[component] = last, current
        components.add(component)
    if force and only_components:
        components.update(only_components)
    component_dependences = {}
    
    if verbose:
        print('all_services:', all_services)
        print('manifest_deployers_befores:', common.manifest_deployers_befores.keys())
        print('*'*80)
        print('all components:', components)
    
    all_components = set(common.all_satchels)
    if only_components and not all_components.issuperset(only_components):
        unknown_components = set(only_components).difference(all_components)
        raise Exception('Unknown components: %s' % ', '.join(sorted(unknown_components)))

    for _c in components:
        if verbose:
            print('checking:', _c)
        deps = set(common.manifest_deployers_befores.get(_c, []))
        if verbose:
            print('deps0:', deps)
        deps = deps.intersection(components)
        if verbose:
            print('deps1:', deps)
        component_dependences[_c] = deps
        
    if verbose:
        print('dependencies:')
        for _c in component_dependences:
            print(_c, component_dependences[_c])
        
    components = list(common.topological_sort(component_dependences.items()))
    plan_funcs = list(get_deploy_funcs(components))
    if components and plan_funcs:
        print('These components have changed:\n')
        for component in sorted(components):
            print((' '*4)+component)
        print('\nDeployment plan:\n')
        for func_name, _ in plan_funcs:
            print(success((' '*4)+func_name))
    else:
        print('Nothing to do!')
        return False
    
    # Execute plan. 
    if preview:
        print('\nTo execute this plan on all hosts run:\n\n    fab %s deploy.run' % env.ROLE)
        return components, plan_funcs
    else:
        with open('/tmp/burlap.progress', 'w') as fout:
            print('%s Beginning plan execution!' % (datetime.datetime.now(),), file=fout)
            fout.flush()
            for func_name, plan_func in plan_funcs:
                print('%s Executing step %s...' % (datetime.datetime.now(), func_name))
                print('%s Executing step %s...' % (datetime.datetime.now(), func_name), file=fout)
                fout.flush()
                if callable(plan_func):
                    plan_func()
                    
                    # Record this step complete.
                    if not only_components:
                        try:
                            thumbprint(components=func_name.split('.')[0])
                        except AssertionError:
                            # On new installs where the host is not yet present, this may fail.
                            pass
                        
                print('%s Done!' % (datetime.datetime.now(),), file=fout)
                fout.flush()
            print('%s Plan execution complete!' % (datetime.datetime.now(),), file=fout)
            fout.flush()
    
    # Create thumbprint.
    if enable_plans and not preview and not common.get_dryrun():
        plan = Plan.get_or_create_next(last_plan=last_plan)
        plan.record_thumbprint(only_components=only_components)

@task_or_dryrun
def run(*args, **kwargs):
    """
    Performs a full deployment.
    
    Parameters:
    
        components := name of satchel to limit deployment to
    """
    from burlap import notifier
    from burlap.common import all_satchels, get_satchel
    
    service = get_satchel('service')
    
    assume_yes = int(kwargs.pop('assume_yes', 0)) or int(kwargs.pop('yes', 0))
    fake = int(kwargs.get('fake', 0))
    
    # Allow satchels to configure connection parameters before we try contacting the hosts.
    #TODO:support ordering?
    services_set = set(env.services)
    for name, satchel in all_satchels.iteritems():

        # Skip satchels that we're not using.
        if name.lower() not in services_set:
            continue

        if hasattr(satchel, 'deploy_pre_run'):
            try:
                satchel.deploy_pre_run()
            except Exception as e:
                print('Error running %s pre-runner: %s' % (name, e))
    
    if env.host_string == env.hosts[0]:
        pending = preview(*args, **kwargs)
        if pending:
            # There are changes that need to be deployed, but confirm first with user.
            if not assume_yes \
            and not raw_input('\nBegin deployment? [yn] ').strip().lower().startswith('y'):
                sys.exit(1)
        else:
            # There are no changes pending, so abort all further tasks.
            sys.exit(1)
    
    if not fake:
        service.pre_deploy()
        
    kwargs['check_outstanding'] = 0
    auto(*args, **kwargs)
    
    if not fake:
        service.post_deploy()
        notifier.notify_post_deployment()

@task_or_dryrun
def test_remotefile():
    f = RemoteFile('/var/log/auth.log')
    f.read()
    print(id(f))
    print('-'*80)
    f = RemoteFile('/var/log/auth.log')
    print(id(f))
    


"""
Disk Tools
==========
"""

import re

from fabric.api import hide, settings, abort

from burlap.utils import run_as_root


def partitions(device=""):
    """
    Get a partition list for all disk or for selected device only

    Example::

        from burlap.disk import partitions

        spart = {'Linux': 0x83, 'Swap': 0x82}
        parts = partitions()
        # parts =  {'/dev/sda1': 131, '/dev/sda2': 130, '/dev/sda3': 131}
        r = parts['/dev/sda1'] == spart['Linux']
        r = r and parts['/dev/sda2'] == spart['Swap']
        if r:
            print("You can format these partitions")
    """
    partitions_list = {}
    with settings(hide('running', 'stdout')):
        res = run_as_root('sfdisk -d %(device)s' % locals())

        spart = re.compile(r'(?P<pname>^/.*) : .* Id=(?P<ptypeid>[0-9a-z]+)')
        for line in res.splitlines():
            m = spart.search(line)
            if m:
                partitions_list[m.group('pname')] = int(m.group('ptypeid'), 16)

    return partitions_list


def getdevice_by_uuid(uuid):
    """
    Get a HDD device by uuid

    Example::

        from burlap.disk import getdevice_by_uuid

        device = getdevice_by_uuid("356fafdc-21d5-408e-a3e9-2b3f32cb2a8c")
        if device:
            mount(device,'/mountpoint')
    """
    with settings(hide('running', 'warnings', 'stdout'), warn_only=True):
        res = run_as_root('blkid -U %s' % uuid)

        if not res.succeeded:
            return None

        return res


def mount(device, mountpoint):
    """
    Mount a partition

    Example::

        from burlap.disk import mount

        mount('/dev/sdb1', '/mnt/usb_drive')
    """
    if not ismounted(device):
        run_as_root('mount %(device)s %(mountpoint)s' % locals())


def swapon(device):
    """
    Active swap partition

    Example::

        from burlap.disk import swapon

        swapon('/dev/sda1')
    """
    if not ismounted(device):
        run_as_root('swapon %(device)s' % locals())


def ismounted(device):
    """
    Check if partition is mounted

    Example::

        from burlap.disk import ismounted

        if ismounted('/dev/sda1'):
           print ("disk sda1 is mounted")
    """
    # Check filesystem
    with settings(hide('running', 'stdout')):
        res = run_as_root('mount')
    for line in res.splitlines():
        fields = line.split()
        if fields[0] == device:
            return True

    # Check swap
    with settings(hide('running', 'stdout')):
        res = run_as_root('swapon -s')
    for line in res.splitlines():
        fields = line.split()
        if fields[0] == device:
            return True

    return False


def mkfs(device, ftype):
    """
    Format filesystem

    Example::

        from burlap.disk import mkfs

        mkfs('/dev/sda2', 'ext4')
    """
    if not ismounted('%(device)s' % locals()):
        run_as_root('mkfs.%(ftype)s %(device)s' % locals())
    else:
        abort("Partition is mounted")


def mkswap(device):
    """
    Format swap partition

    Example::

        from burlap.disk import mkswap

        mkswap('/dev/sda2')
    """
    if not ismounted(device):
        run_as_root('mkswap %(device)s' % locals())
    else:
        abort("swap partition is mounted")



from __future__ import print_function

import os
import re
import shutil
import getpass
from commands import getstatusoutput

from fabric.contrib.files import append
from fabric.api import settings

import burlap
from burlap.common import env
from burlap.tests.base import TestCase
from burlap.project import project
from burlap.context import set_cwd
from burlap.deploy import STORAGE_LOCAL

class DjTests(TestCase):
    
    def getstatusoutput(self, cmd):
        print(cmd)
        status, output = getstatusoutput(cmd)
        print('output:', output)
        return status, output
    
    def test_migrate(self):
        
        burlap_dir = os.path.abspath(os.path.split(burlap.__file__)[0])
        
        d = '/tmp/test_dj_migrate'
        if os.path.isdir(d):
            shutil.rmtree(d)
        os.makedirs(d)
        
        activate_cmd = '. {d}/.env/bin/activate;'.format(d=d)
        with set_cwd(d):
            project.create_skeleton(
                project_name='test_dj_migrate',
                roles='prod',
                components='dj',
            )
            
            assert not os.path.isfile('/tmp/test_dj_migrate/plans/prod/000/thumbprints/test-dj-migrate-1')
            assert not os.path.isfile('/tmp/test_dj_migrate/plans/prod/000/thumbprints/test-dj-migrate-2')
            
            # Simulate multiple remote hosts my creating aliases of localhost.
            # Note, for testing this on your localhost for a user without passwordless sudo,
            # you may have to run: `sudo chmod 777 /etc/hosts`
            # This won't work on Travis, where these will instead be set in .travis.yml.
            print('Modifying /etc/hosts...')
            env.host_string = 'localhost'
            env.hosts = [env.host_string]
            env.user = getpass.getuser()
            with settings(warn_only=True):
                for use_sudo in (False, True):
                    ret = append(filename='/etc/hosts', text='127.0.0.1 test-dj-migrate-1\n127.0.0.1 test-dj-migrate-2', use_sudo=use_sudo)
                    if ret is None:
                        break
            
            os.system('ln -s %s %s/' % (burlap_dir, d))
            
            project.update_settings({
                    'plan_storage': STORAGE_LOCAL,
                    'plan_data_dir': os.path.join(d, 'plans'),
                    'services': ['dj'],
                    'dj_settings_module': 'test_dj_migrate.settings',
                    'default_site': 'testsite1',
                    'default_role': 'prod',
                    'dj_local_project_dir': 'src',
                    'dj_project_dir': '%s/src' % d,
                    'dj_manage_media': False,
                    'dj_manage_migrations': True,
                    'dj_manage_cmd': '%s/.env/bin/python manage.py' % d,
                    'dj_version': [1, 10, 6],
                    # This is necessary to stop get_current_hostname() from attempting to lookup our actual hostname.
                    '_ip_to_hostname': {
                        'test-dj-migrate-1': 'test-dj-migrate-1',
                        'test-dj-migrate-2': 'test-dj-migrate-2',
                    },
                },
                role='all')
                
            project.update_settings({
                    'hosts': ['test-dj-migrate-1', 'test-dj-migrate-2'],
                    'available_sites_by_host':{
                        'test-dj-migrate-1': [
                            'testsite1',
                        ],
                        'test-dj-migrate-2': [
                            'testsite2',
                        ]
                    },
                    'sites': {
                        'testsite1': {
                            'apache_domain_template': 'testsite1.test-dj-migrate-1.com',
                        },
                        'testsite2': {
                            'apache_domain_template': 'testsite2.test-dj-migrate-2.com',
                        },
                    },
                },
                role='prod')
            
            # Migrate built-in apps.
            kwargs = dict(
                activate_cmd=activate_cmd,
            )
            status, output = self.getstatusoutput('{activate_cmd} fab prod deploy.run:yes=1'.format(**kwargs))
            assert not status
            # The migrations should have been run on both hosts.
            assert '[test-dj-migrate-1] run:' in output
            assert '[test-dj-migrate-2] run:' in output
            assert os.path.isfile('/tmp/test_dj_migrate/plans/prod/000/thumbprints/test-dj-migrate-1')
            assert os.path.isfile('/tmp/test_dj_migrate/plans/prod/000/thumbprints/test-dj-migrate-2')
            
            # Create custom app.
            kwargs = dict(
                src_dir='/tmp/test_dj_migrate/src',
                activate_cmd=activate_cmd,
            )
            status, output = self.getstatusoutput('{activate_cmd} cd {src_dir}; python manage.py startapp myapp'.format(**kwargs))
            assert os.path.isdir('/tmp/test_dj_migrate/src/myapp')
            
            # Add myapp to installed apps list.
            settings_fn = '/tmp/test_dj_migrate/src/test_dj_migrate/settings.py'
            with open(settings_fn) as fin:
                settings_text = fin.read()
            p = re.compile(r'INSTALLED_APPS = \[([^\]]+)', flags=re.MULTILINE)
            settings_text = p.sub(r"INSTALLED_APPS = [\1    'myapp',\n", settings_text)
            with open(settings_fn, 'w') as fout:
                fout.write(settings_text)
            status, output = self.getstatusoutput('cd {src_dir}/test_dj_migrate; rm -f *.pyc'.format(**kwargs))
            assert not status
            
            # Populate model.
            open('/tmp/test_dj_migrate/src/myapp/models.py', 'w').write('''
from __future__ import unicode_literals

from django.db import models

class MyModel(models.Model):
    pass

''')
            kwargs = dict(
                src_dir='/tmp/test_dj_migrate/src',
                activate_cmd=activate_cmd,
            )
            status, output = self.getstatusoutput('{activate_cmd} cd {src_dir}; python manage.py makemigrations myapp'.format(**kwargs))
            assert os.path.isfile('/tmp/test_dj_migrate/src/myapp/migrations/0001_initial.py')

            # Apply migrations.
            kwargs = dict(
                activate_cmd=activate_cmd,
            )
            #cmd = '{activate_cmd} fab prod deploy.show_diff'.format(activate_cmd=activate_cmd)
            #cmd = '{activate_cmd} fab prod deploy.run'.format(activate_cmd=activate_cmd)
            #status, output = self.getstatusoutput('{activate_cmd} fab prod dj.configure:dryrun=1,verbose=1'.format(**kwargs))
            status, output = self.getstatusoutput('{activate_cmd} fab prod deploy.run:yes=1'.format(**kwargs))
            assert not status
            # The migrations should have been run on both hosts.
            assert ('test-dj-migrate-1] run: export SITE=testsite1; export ROLE=prod; cd /tmp/test_dj_migrate/src; '
                '/tmp/test_dj_migrate/.env/bin/python manage.py migrate') in output
            assert ('test-dj-migrate-2] run: export SITE=testsite2; export ROLE=prod; cd /tmp/test_dj_migrate/src; '
                '/tmp/test_dj_migrate/.env/bin/python manage.py migrate') in output

            assert os.path.isfile('/tmp/test_dj_migrate/plans/prod/001/thumbprints/test-dj-migrate-1')
            assert os.path.isfile('/tmp/test_dj_migrate/plans/prod/001/thumbprints/test-dj-migrate-2')



from __future__ import print_function

import os
import sys
import subprocess
from pprint import pprint

from fabric.api import env#, runs_once

from burlap.constants import *
from burlap import ServiceSatchel
from burlap.common import LocalRenderer, pretty_bytes
from burlap.decorators import task, runs_once

CONNECTION_HANDLER_DJANGO = 'django'

class DatabaseSatchel(ServiceSatchel):
    
    name = 'db'
    
    _database_renderers = {} # {(name, site): renderer}
        
    def set_defaults(self):
            
        # Local cache for renderers.
        self._database_renderers = {} # {(name, site): renderer}
                
        # If set, allows remote users to connect to the database.
        # This shouldn't be necessary if the webserver and database
        # share the same server.
        self.env.allow_remote_connections = False
        
        # Directory where database snapshots will be temporarily stored.
        self.env.dump_dest_dir = '/tmp'
        
        self.env.dump_archive_dir = 'snapshots'
        
        # Default filename of database snapshots.
        self.env.dump_fn_template = '{dump_dest_dir}/db_{db_type}_{SITE}_{ROLE}_{db_name}_$(date +%Y%m%d).sql.gz'
        
        # This overrides the built-in dump command.
        self.env.dump_command = None
         
        # This overrides the built-in load command.
        self.env.load_command = None
        
        # {hostname: {username: ?, password: ?}}
        self.env.root_logins = {}
        
        # Settings for specific databases within the server.
        self.env.databases = {} # {name: {}}
        
        self.env.default_db_name = 'default'

    def get_database_defaults(self):
        """
        Returns a dictionary of default settings for each database.
        """
        return dict(
            # {name: None=burlap settings, Django=Django Python settings}
            connection_handler=None,
        )
        
    @task
    def execute(self, sql, name='default', site=None, **kwargs):
        raise NotImplementedError
    @task
    def execute_file(self, filename, name='default', site=None, **kwargs):
        raise NotImplementedError

    @task
    def set_root_login(self, r):
        """
        Looks up the root login for the given database on the given host and sets
        it to environment variables.
        
        Populates these standard variables:
        
            db_root_password
            db_root_username
            
        """
        
        # Check the legacy password location.
        try:
            r.env.db_root_username = r.env.root_username
        except AttributeError:
            pass
        try:
            r.env.db_root_password = r.env.root_password
        except AttributeError:
            pass
        
        # Check the new password location.
        key = r.env.get('db_host')
        if self.verbose:
            print('root login key:', key)
#         print('r.env.root_logins:', r.env.root_logins)
        if key in r.env.root_logins:
            data = r.env.root_logins[key]
#             print('data:', data)
            if 'username' in data:
                r.env.db_root_username = data['username']
                r.genv.db_root_username = data['username']
            if 'password' in data:
                r.env.db_root_password = data['password']
                r.genv.db_root_password = data['password']
        else:
            msg = 'Warning: No root login entry found for host %s in role %s.' % (r.env.get('db_host'), self.genv.get('ROLE'))
            print(msg, file=sys.stderr)
            #warnings.warn(msg, UserWarning)

    def database_renderer(self, name=None, site=None, role=None):
        """
        Renders local settings for a specific database.
        """
        
        name = name or self.env.default_db_name
        
        site = site or self.genv.SITE
        
        role = role or self.genv.ROLE
        
        key = (name, site, role)
        self.vprint('checking key:', key)
        if key not in self._database_renderers:
            self.vprint('No cached db renderer, generating...')
            
            d = type(self.genv)(self.lenv)
            d.update(self.get_database_defaults())
            d.update(self.env.databases.get(name, {}))
            d['db_name'] = name
            
            if d.connection_handler == CONNECTION_HANDLER_DJANGO:
                self.vprint('Using django handler...')
                dj = self.get_satchel('dj')
#                 _d = type(self.genv)()
                if self.verbose:
                    print('Loading Django DB settings for site {} and role {}.'.format(site, role), file=sys.stderr)
                dj.set_db(name=name, site=site, role=role)
                _d = dj.local_renderer.collect_genv(include_local=True, include_global=False)

                # Copy "dj_db_*" into "db_*".
                for k, v in _d.items():
                    if k.startswith('dj_db_'):
                        _d[k[3:]] = v
                    del _d[k]

                if self.verbose:
                    print('Loaded:')
                    pprint(_d)
                d.update(_d)
            
            r = LocalRenderer(self, lenv=d)
            
            # Optionally set any root logins needed for administrative commands.
            self.set_root_login(r)
            
            self._database_renderers[key] = r
        else:
            self.vprint('Cached db renderer found.')
        
        return self._database_renderers[key]
        
    @task
    def configure(self, *args, **kwargs):
        raise NotImplementedError
    
    @task
    def get_free_space(self):
        """
        Return free space in bytes.
        """
        cmd = "df -k | grep -vE '^Filesystem|tmpfs|cdrom|none|udev|cgroup' | awk '{ print($1 \" \" $4 }'"
        lines = [_ for _ in self.run(cmd).strip().split('\n') if _.startswith('/')]
        assert len(lines) == 1, 'Ambiguous devices: %s' % str(lines)
        device, kb = lines[0].split(' ')
        free_space = int(kb) * 1024
        self.vprint('free_space (bytes):', free_space)
        return free_space
    
    @task
    def get_size(self):
        """
        Retrieves the size of the database in bytes.
        """
        #TODO:remove django hardcoding
#         dj = self.get_satchel('dj')
#         dj.set_db(site=env.SITE, role=env.ROLE)
#         r = self.local_renderer
#         if 'postgres' in self.genv.db_engine or 'postgis' in self.genv.db_engine:
#             output = r.run('psql --user=%(db_postgresql_postgres_user)s --tuples-only -c "SELECT pg_database_size(\'%(db_name)s\');"')
#             output = self.run(cmd)
#             output = int(output.strip().split('\n')[-1].strip())
#             self.vprint('database size (bytes):', output)
#             return output
#         else:
        raise NotImplementedError

    @task
    def load_table(self, table_name, src, dst='localhost', name=None, site=None):
        """
        Directly transfers a table between two databases.
        """
        raise NotImplementedError

    @task
    def load_db_set(self, name, r=None):
        """
        Loads database parameters from a specific named set.
        """
        r = r or self
        db_set = r.genv.db_sets.get(name, {})
        r.genv.update(db_set)

    @task
    def loadable(self, src, dst):
        """
        Determines if there's enough space to load the target database.
        """
        from fabric import state
        from fabric.task_utils import crawl
        
        src_task = crawl(src, state.commands)
        assert src_task, 'Unknown source role: %s' % src
        
        dst_task = crawl(dst, state.commands)
        assert dst_task, 'Unknown destination role: %s' % src
        
        # Get source database size.
        src_task()
        env.host_string = env.hosts[0]
        src_size_bytes = self.get_size()
        
        # Get target database size, if any.
        dst_task()
        env.host_string = env.hosts[0]
        try:
            dst_size_bytes = self.get_size()
        except (ValueError, TypeError):
            dst_size_bytes = 0
        
        # Get target host disk size.
        free_space_bytes = self.get_free_space()
        
        # Deduct existing database size, because we'll be deleting it.
        balance_bytes = free_space_bytes + dst_size_bytes - src_size_bytes
        balance_bytes_scaled, units = pretty_bytes(balance_bytes)
        
        viable = balance_bytes >= 0
        if self.verbose:
            print('src_db_size:', pretty_bytes(src_size_bytes))
            print('dst_db_size:', pretty_bytes(dst_size_bytes))
            print('dst_free_space:', pretty_bytes(free_space_bytes))
            print
            if viable:
                print('Viable! There will be %.02f %s of disk space left.' % (balance_bytes_scaled, units))
            else:
                print('Not viable! We would be %.02f %s short.' % (balance_bytes_scaled, units))
        
        return viable
    
    @task
    def dumpload(self, site=None, role=None):
        """
        Dumps and loads a database snapshot simultaneously.
        Requires that the destination server has direct database access
        to the source server.
        
        This is better than a serial dump+load when:
        1. The network connection is reliable.
        2. You don't need to save the dump file.
        
        The benefits of this over a dump+load are:
        1. Usually runs faster, since the load and dump happen in parallel.
        2. Usually takes up less disk space since no separate dump file is
            downloaded.
        """
        raise NotImplementedError

    def render_fn(self, fn):
        return subprocess.check_output('echo %s' % fn, shell=True)
        
    def get_default_db_fn(self, fn_template=None, dest_dir=None, name=None, site=None):
        
        r = self.database_renderer(name=name, site=site)
        r.dump_dest_dir = dest_dir
        
        fn = r.format(fn_template or r.env.dump_fn_template)
        fn = self.render_fn(fn)
        fn = fn.strip()
        return fn
        
    @task
    @runs_once
    def dump(self, dest_dir=None, to_local=1, from_local=0, archive=0, dump_fn=None, name=None, site=None, use_sudo=0, cleanup=1):
        """
        Exports the target database to a single transportable file on the localhost,
        appropriate for loading using load().
        """
        r = self.database_renderer(name=name, site=site)

        use_sudo = int(use_sudo)
        
        from_local = int(from_local)
        
        to_local = int(to_local)
        
        dump_fn = dump_fn or r.env.dump_fn_template
        
        # Render the snapshot filename.
        r.env.dump_fn = self.get_default_db_fn(
            fn_template=dump_fn,
            dest_dir=dest_dir,
            name=name,
            site=site,
        )
        
        # Dump the database to a snapshot file.
        #if not os.path.isfile(os.path.abspath(r.env.dump_fn))):
        r.pc('Dumping database snapshot.')
        if from_local:
            r.local(r.env.dump_command)
        elif use_sudo:
            r.sudo(r.env.dump_command)
        else:
            r.run(r.env.dump_command)
        
        # Download the database dump file on the remote host to localhost.
        if not from_local and to_local:
            r.pc('Downloading database snapshot to localhost.')
            r.local('rsync -rvz --progress --recursive --no-p --no-g '
                '--rsh "ssh -o StrictHostKeyChecking=no -i {key_filename}" {user}@{host_string}:{dump_fn} {dump_fn}')
            
            # Delete the snapshot file on the remote system.
            if int(cleanup):
                r.pc('Deleting database snapshot on remote host.')
                r.sudo('rm {dump_fn}')
        
        # Move the database snapshot to an archive directory.
        if to_local and int(archive):
            r.pc('Archiving database snapshot.')
            db_fn = r.render_fn(r.env.dump_fn)
            r.env.archive_fn = '%s/%s' % (env.db_dump_archive_dir, os.path.split(db_fn)[-1])
            r.local('mv %s %s' % (db_fn, env.archive_fn))
        
        return r.env.dump_fn

    @task
    @runs_once
    def load(self, db_dump_fn='', prep_only=0, force_upload=0, from_local=0):
        """
        Restores a database snapshot onto the target database server.
        
        If prep_only=1, commands for preparing the load will be generated,
        but not the command to finally load the snapshot.
        """
        raise NotImplementedError
        
    @task
    def shell(self, name='default', user=None, password=None, root=0, verbose=1, write_password=1, no_db=0, no_pw=0):
        """
        Opens a SQL shell to the given database, assuming the configured database
        and user supports this feature.
        """
        raise NotImplementedError

    @task
    def create(self, **kwargs):
        """
        Creates the target database.
        """
        raise NotImplementedError
        
    @task
    def drop_views(self, name=None, site=None):
        """
        Drops all views.
        """
        raise NotImplementedError

    @task
    def exists(self, name='default', site=None):
        """
        Returns true if a database with the given name exists. False otherwise.
        """
        raise NotImplementedError

#db = DatabaseSatchel()



from __future__ import print_function

from fabric.contrib.files import exists
from fabric.context_managers import cd
from fabric.api import run
 
from burlap.common import set_verbose
from burlap.git import git
from burlap.tests.functional_tests.base import TestCase
from burlap.deploy import thumbprint, clear_fs_cache, delete_plan_data_dir
#from burlap.packager import packager

class GitTests(TestCase):

    def test_hooks(self):

        set_verbose(True)
        git.genv.ROLE = 'prod'
        git.genv.services = ['git']
        git.clear_caches()

        print('Installing git...')
        git.install_packages() # fails on Ubuntu 14 under Travis-CI?

        print('Setting up sample git repo...')
        run('mkdir /tmp/mygithookrepo')
        with cd('/tmp/mygithookrepo'):
            run('git init')
        assert not exists('/tmp/mygithookrepo/.git/hooks/post-checkout')

        print('Configuring git...')
        git.env.enabled = True
        git.env.hooks = {'/tmp/mygithookrepo': ['git/post-checkout']}
        git.clear_local_renderer()
        clear_fs_cache()
        delete_plan_data_dir()
        cm = git.current_manifest
        print('cm1:', cm)
        assert 'hooks' in cm

        added_hooks, removed_hooks = git.get_changed_hooks()
        print('added_hooks:', added_hooks)
        assert added_hooks == {'/tmp/mygithookrepo': ['git/post-checkout']}
        print('removed_hooks:', removed_hooks)
        assert removed_hooks == {}

        print('Installing git hooks...')
        git.clear_local_renderer()
        git.configure()
        clear_fs_cache()
        delete_plan_data_dir()
        print('-'*80)
        print('Thumbprinting...')
        thumbprint(components=git.name)
        clear_fs_cache()
        print('-'*80)

        assert exists('/tmp/mygithookrepo/.git/hooks/post-checkout')



from __future__ import print_function

from burlap import Satchel
from burlap.constants import *
from burlap.decorators import task

#http://askubuntu.com/a/555366/13217
class VirtualboxSatchel(Satchel):
    
    name = 'virtualbox'

    @task(precursors=['packager'])
    def configure(self):
        """
        Enables the repository for a most current version on Debian systems.
        
            https://www.rabbitmq.com/install-debian.html
        """
        
        os_version = self.os_version
        if not self.dryrun and os_version.distro != UBUNTU:
            raise NotImplementedError("OS %s is not supported." % os_version)
        
        r = self.local_renderer
        
        r.env.codename = (r.run('lsb_release -c -s') or "`lsb_release -c -s`").strip()
        r.sudo('apt-add-repository "deb http://download.virtualbox.org/virtualbox/debian {codename} contrib"')
        r.sudo('cd /tmp; wget -q https://www.virtualbox.org/download/oracle_vbox.asc -O- | sudo apt-key add -')
        r.sudo('apt-get update')

virtualbox = VirtualboxSatchel()



from __future__ import print_function

import sys

from burlap import ServiceSatchel
from burlap.constants import *
from burlap.decorators import task 

class CronSatchel(ServiceSatchel):
    
    name = 'cron'
    
    ## Service options.
    
    #ignore_errors = True
        
    post_deploy_command = None
    
    @property
    def packager_system_packages(self):
        return {
            FEDORA: ['crontabs'],
            (UBUNTU, '12.04'): ['cron'],
        }

    def set_defaults(self):
        self.env.enabled = True
        self.env.crontabs_available = type(self.genv)() # {name:[cron lines]}
        self.env.command = 'cron'
        self.env.user = 'www-data'
        self.env.python = None
        self.env.crontab_headers = ['PATH=/usr/sbin:/usr/bin:/sbin:/bin\nSHELL=/bin/bash']
        self.env.stdout_log_template = r'/tmp/chroniker-%(SITE)s-stdout.$(date +\%%d).log'
        self.env.stderr_log_template = r'/tmp/chroniker-%(SITE)s-stderr.$(date +\%%d).log'
        self.env.crontabs_selected = [] # [name]
           
        self.env.service_commands = {
            START:{
                FEDORA: 'systemctl start crond.service',
                UBUNTU: 'service cron start',
            },
            STOP:{
                FEDORA: 'systemctl stop crond.service',
                UBUNTU: 'service cron stop',
            },
            DISABLE:{
                FEDORA: 'systemctl disable crond.service',
                UBUNTU: 'chkconfig cron off',
                (UBUNTU, '14.04'): 'update-rc.d -f cron remove',
            },
            ENABLE:{
                FEDORA: 'systemctl enable crond.service',
                UBUNTU: 'chkconfig cron on',
                (UBUNTU, '14.04'): 'update-rc.d cron defaults',
            },
            RESTART:{
                FEDORA: 'systemctl restart crond.service',
                UBUNTU: 'service cron restart; sleep 3',
            },
            STATUS:{
                FEDORA: 'systemctl status crond.service',
                UBUNTU: 'service cron status',
            },
        }
        
    def render_paths(self):
        r = self.local_renderer
        r.env.cron_stdout_log = r.format(r.env.stdout_log_template)
        r.env.cron_stderr_log = r.format(r.env.stderr_log_template)
    
    def deploy(self, site=None):
        """
        Writes entire crontab to the host.
        """
        r = self.local_renderer
        
        cron_crontabs = []
#         if self.verbose:
#             print('hostname: "%s"' % (hostname,), file=sys.stderr) 
        for site, site_data in self.iter_sites(site=site):
            if self.verbose:
                print('site:', site, file=sys.stderr)
                print('env.crontabs_selected:', self.env.crontabs_selected, file=sys.stderr)
                
            for selected_crontab in self.env.crontabs_selected:
                lines = self.env.crontabs_available.get(selected_crontab, [])
                if self.verbose:
                    print('lines:', lines, file=sys.stderr)
                for line in lines:
                    cron_crontabs.append(r.format(line))
        
        if not cron_crontabs:
            return
        
        cron_crontabs = self.env.crontab_headers + cron_crontabs
        cron_crontabs.append('\n')
        r.env.crontabs_rendered = '\n'.join(cron_crontabs)
        fn = self.write_to_file(content=r.env.crontabs_rendered)
        r.env.put_remote_path = r.put(local_path=fn)
        r.sudo('crontab -u {cron_user} {put_remote_path}')
    
    @task(precursors=['packager', 'user', 'tarball'])
    def configure(self, **kwargs):
        if self.env.enabled:
            kwargs['site'] = ALL
            self.deploy(**kwargs)
            self.enable()
            self.restart()
        else:
            self.disable()
            self.stop()
        
cron = CronSatchel()



import hashlib

from fabric.api import hide, run, settings

import burlap
from burlap.tests.functional_tests.base import TestCase

class Md5Tests(TestCase):
    
    def test_md5sum_empty_file(self):
        try:
            run('touch f1')
            expected_hash = hashlib.md5('').hexdigest()
            assert burlap.files.file.md5sum('f1') == expected_hash
        finally:
            run('rm -f f1')

    def test_md5sum(self):
        try:
            run('echo -n hello > f2')
            expected_hash = hashlib.md5('hello').hexdigest()
            assert burlap.files.file.md5sum('f2') == expected_hash
        finally:
            run('rm -f f2')

    def test_md5sum_not_existing_file(self):
        with settings(hide('warnings')):
            assert burlap.files.file.md5sum('doesnotexist') is None



from __future__ import print_function

from burlap.tests.base import TestCase
from burlap.jirahelper import jirahelper

class JiraHelperTests(TestCase):
    
    def test_jirahelper(self):        
        test_log = """
38fc750991db7df363322e77e4ce56d81d0f7eee Fixed migration conflict
5bf53f847bd29eb796f15715a1cbe8a8437d6847 TICKET-3378 - Batch Edit Disassociation
6c9feb0b550c4b96150b4524d6fbd1ded294bb0f Merged in TICKET-3346 (pull request #932)
d7715a04ee829ac16f61a406bbd0d84e3b80eef2 Merged in TICKET-3469 (pull request #933)
b0dd4fe1ab041472f0d83a421dc8be7c9e9bc604 Fixed django settings.
3f34da3173cc4f07ba14ae5786a5f88023592dd7 Regenerate field migration to fix migration conflict
88867aa2209869106dcb35dc7b1db62dee0f543f Fix EmptyResultSet error raised by empty queries in Django admin
5b5a63b69004fbe11ffda310faee5ff069267c50 Merged in TICKET-3481 (pull request #931)
46ef7818aed657b9bc99c181e6c73a7c6a47a487 TICKET-3481 fixed error creating model
1d4c2dbeab0e04d18f75e11f26ddf41aa3a3f00b Merged in TICKET-3473 (pull request #914)
        """
        jirahelper.env.ticket_pattern = 'TICKET-[0-9]+'
        tickets = sorted(jirahelper.get_tickets_from_str(test_log))
        print('tickets:', tickets)
        assert tickets == ['TICKET-3346', 'TICKET-3378', 'TICKET-3469', 'TICKET-3473', 'TICKET-3481']



from __future__ import print_function

import sys
import re
import traceback
from pprint import pprint

from burlap import ContainerSatchel
from burlap.constants import *
from burlap.decorators import task

class JiraHelperSatchel(ContainerSatchel):
    
    name = 'jirahelper'
    
    def set_defaults(self):
    
        self.env.server = None
        self.env.basic_auth_username = None
        self.env.basic_auth_password = None
        self.env.update_from_git = False
        self.env.ticket_update_message_template = 'This has been deployed to {role}.'
        
        # A map of status->transition to follow when making deployments.
        self.env.deploy_workflow = {}
        
        # A map of the new-status->new-assignee to auto-assign.
        self.env.assignee_by_status = {}
        
        # The regex used to search the commit messages for a ticket number.
        self.env.ticket_pattern = None

    def get_tickets_from_str(self, s):
        pattern = re.compile(self.env.ticket_pattern, flags=re.I)
        tickets = []
        if self.env.ticket_pattern:
            pattern = re.compile(self.env.ticket_pattern, flags=re.I)
            tickets.extend(pattern.findall(s))
        return set(_.strip().upper() for _ in tickets)

    @task
    def get_tickets_between_commits(self, a, b):
        from burlap.git import gittracker
        tickets = []
        if self.env.ticket_pattern:
            ret = gittracker.get_logs_between_commits(a, b)
            tickets.extend(self.get_tickets_from_str(ret))
        self.vprint(tickets)
        return tickets
    
    @task
    def update_tickets_from_git(self, from_commit=None, to_commit=None):
        """
        Run during a deployment.
        Looks at all commits between now and the last deployment.
        Finds all ticket numbers and updates their status in Jira.
        """
        from jira import JIRA, JIRAError
        from burlap.deploy import get_last_current_diffs
        from burlap.git import gittracker, CURRENT_COMMIT
        
        r = self.local_renderer
        
#         get_current_commit = gittracker.get_current_commit
#         GITTRACKER = gittracker.name.upper()
        
        # Ensure this is only run once per role.
        if self.genv.host_string != self.genv.hosts[-1]:
            self.vprint('Not first server. Aborting.')
            return
        
        print('self.env.update_from_git:', self.env.update_from_git)
        print('self.genv.jirahelper_update_from_git:', self.genv.jirahelper_update_from_git)
        if not self.env.update_from_git:
            self.vprint('Update from git disabled. Aborting.')
            return
        
        if not self.env.ticket_pattern:
            self.vprint('No ticket pattern defined. Aborting.')
            return
        
        if not self.env.basic_auth_username or not self.env.basic_auth_password:
            self.vprint('Username or password not given. Aborting.')
            return
        
        # During a deployment, we should be given these, but for testing,
        # lookup the diffs dynamically.
        last = gittracker.last_manifest
        current = gittracker.current_manifest
        
        last_commit = from_commit or last.current_commit#[CURRENT_COMMIT]
        print('last_commit:', last_commit)
        current_commit = to_commit or current[CURRENT_COMMIT]
        print('current_commit:', current_commit)
        
        if not last_commit or not current_commit:
            print('Missing commit ID. Aborting.')
            return
        
        if self.verbose:
            print('-'*80)
            print('last.keys:', last.keys())
            print('-'*80)
            print('current.keys:', current.keys())
        
#         try:
#             last_commit = last['GITTRACKER']['current_commit']
#         except KeyError:
#             return
#         current_commit = current['GITTRACKER']['current_commit']
        
        # Find all tickets deployed between last deployment and now.
        tickets = self.get_tickets_between_commits(current_commit, last_commit)
        if self.verbose:
            print('tickets:', tickets)
        
        # Update all tickets in Jira.
        jira = JIRA({
            'server': self.env.server
        }, basic_auth=(self.env.basic_auth_username, self.env.basic_auth_password))
        for ticket in tickets:
            
            # Mention this Jira updated.
            r.env.role = r.genv.ROLE.lower()
            comment = r.format(self.env.ticket_update_message_template)
            print('Commenting on ticket %s: %s' % (ticket, comment))
            if not self.dryrun:
                jira.add_comment(ticket, comment) 
            
            # Update ticket status.
            recheck = False
            while 1:
                print('Looking up jira ticket %s...' % ticket)
                issue = jira.issue(ticket)
                print('Ticket %s retrieved.' % ticket)
                transition_to_id = dict((t['name'], t['id']) for t in jira.transitions(issue))
                print('%i allowable transitions found:')
                pprint(transition_to_id)
                print('issue.fields.status.id:', issue.fields.status.id)
                print('issue.fields.status.name:', issue.fields.status.name)
                jira_status_id = issue.fields.status.name.title()
                print('jira_status_id:', jira_status_id)
                next_transition_name = self.env.deploy_workflow.get(jira_status_id)
                print('next_transition_name:', next_transition_name)
                next_transition_id = transition_to_id.get(next_transition_name)
                print('next_transition_id:', next_transition_id)
                if next_transition_name:
                    new_fields = {}
                    new_assignee = self.env.assignee_by_status.get(
                        #issue.fields.status.name.title(),
                        next_transition_name,
                        issue.fields.assignee.name,
                    )
                    if new_assignee == 'reporter':
                        new_assignee = issue.fields.reporter.name
    #                 print('new_assignee:', new_assignee)
                        
                    print('Updating ticket %s to status %s (%s) and assigning it to %s.' \
                        % (ticket, next_transition_name, next_transition_id, new_assignee))
                    if not self.dryrun:
                        try:
                            jira.transition_issue(
                                issue,
                                next_transition_id,
                            )
                            recheck = True
                        except AttributeError as e:
                            print('Unable to transition ticket %s to %s: %s' \
                                % (ticket, next_transition_name, e), file=sys.stderr)
                            traceback.print_exc()
                        
                        # Note assignment should happen after transition, since the assignment may
                        # effect remove transitions that we need.
                        try:
                            if new_assignee:
                                print('Assigning ticket %s to %s.' % (ticket, new_assignee))
                                jira.assign_issue(issue, new_assignee)
                            else:
                                print('No new assignee found.')
                        except JIRAError as e:
                            print('Unable to reassign ticket %s to %s: %s' \
                                % (ticket, new_assignee, e), file=sys.stderr)
                else:
                    recheck = False
                    print('No transitions found for ticket %s currently in status "%s".' \
                        % (ticket, issue.fields.status.name))
                        
                if not recheck:
                    break

jirahelper = JiraHelperSatchel()



from __future__ import print_function

from burlap import ServiceSatchel
from burlap.constants import *
from burlap.decorators import task

class BluetoothSatchel(ServiceSatchel):
    
    name = 'bluetooth'
    
    @property
    def packager_system_packages(self):
        return {
            UBUNTU: ['bluetooth', 'bluez', 'python-bluez' 'bluez-firmware', 'blueman', 'pi-bluetooth'],
            DEBIAN: ['bluetooth', 'bluez', 'python-bluez', 'bluez-firmware', 'blueman', 'pi-bluetooth'],
        }
    
    def set_defaults(self):
        pass
    
    def scan(self):
        r = self.local_renderer
        #r.sudo('hciconfig hci0 piscan')
        r.run('hcitool scan')
    
    def add(self, name):
        r = self.local_renderer
        r.sudo('hciconfig hci0 name "%s"' % name)
    
    @task(precursors=['packager', 'user'])
    def configure(self):
        if self.env.enabled:
            self.install_packages()
            r = self.local_renderer
            #http://blog.davidvassallo.me/2014/05/11/android-linux-raspberry-pi-bluetooth-communication/
            r.sudo('echo "DisablePlugins = pnat" >> /etc/bluetooth/main.conf')

            #TODO:fix bluetooth.btcommon.BluetoothError: (2, 'No such file or directory')
            #https://www.raspberrypi.org/forums/viewtopic.php?f=63&t=133263
            #sudo nano /lib/systemd/system/bluetooth.service
            #-ExecStart=/usr/lib/bluetooth/bluetoothd            
            #+ExecStart=/usr/lib/bluetooth/bluetoothd -C
            #sudo sdptool add SP
            
            self.reboot()

BluetoothSatchel()



"""
OpenSSH tasks
=============

This module provides tools to manage OpenSSH server and client.

"""
from __future__ import print_function

# from fabric.api import hide, shell_env
# from fabric.contrib.files import append, sed

# from burlap.service import is_running, restart
# from burlap.files import watch

from burlap import Satchel
from burlap.constants import *
from burlap.decorators import task

# def harden(allow_root_login=False, allow_password_auth=False,
#            sshd_config='/etc/ssh/sshd_config'):
#     """
#     Apply best practices for ssh security.
# 
#     See :func:`burlap.ssh.disable_password_auth` and
#     :func:`burlap.ssh.disable_root_login` for a detailed
#     description.
# 
#     ::
# 
#         import burlap
# 
#         # This will apply all hardening techniques.
#         burlap.ssh.harden()
# 
#         # Only apply some of the techniques.
#         burlap.ssh.harden(allow_password_auth=True)
# 
#         # Override the sshd_config file location.
#         burlap.ssh.harden(sshd_config='/etc/sshd_config')
# 
#     """
# 
#     if not allow_password_auth:
#         disable_password_auth(sshd_config=sshd_config)
# 
#     if not allow_root_login:
#         disable_root_login(sshd_config=sshd_config)
# 
# 
# def disable_password_auth(sshd_config='/etc/ssh/sshd_config'):
#     """
#     Do not allow users to use passwords to login via ssh.
#     """
# 
#     _update_ssh_setting(sshd_config, 'PasswordAuthentication', 'no')
# 
# 
# def enable_password_auth(sshd_config='/etc/ssh/sshd_config'):
#     """
#     Allow users to use passwords to login via ssh.
#     """
# 
#     _update_ssh_setting(sshd_config, 'PasswordAuthentication', 'yes')
# 
# 
# def disable_root_login(sshd_config='/etc/ssh/sshd_config'):
#     """
#     Do not allow root to login via ssh.
#     """
# 
#     _update_ssh_setting(sshd_config, 'PermitRootLogin', 'no')
# 
# 
# def enable_root_login(sshd_config='/etc/ssh/sshd_config'):
#     """
#     Allow root to login via ssh.
#     """
# 
#     _update_ssh_setting(sshd_config, 'PermitRootLogin', 'yes')
# 
# 
# def _update_ssh_setting(sshd_config, name, value):
#     """
#     Update a yes/no setting in the SSH config file
#     """
# 
#     with watch(sshd_config) as config_file:
# 
#         with shell_env():
# 
#             # First try to change existing setting
#             sed(sshd_config,
#                 r'^(\s*#\s*)?%s\s+(yes|no)' % name,
#                 '%s %s' % (name, value),
#                 use_sudo=True)
# 
#             # Then append setting if it's still missing
#             _append(sshd_config,
#                     '%s %s' % (name, value),
#                     use_sudo=True)
# 
#     if config_file.changed and is_running('ssh'):
#         restart('ssh')


# def _append(filename, regex, use_sudo):
#     """
#     Less verbose append
#     """
#     with hide('stdout', 'warnings'):
#         return append(filename, regex, use_sudo=use_sudo)

class SSHNiceSatchel(Satchel):

    name = 'sshnice'
    
    @property
    def packager_system_packages(self):
        return {
            FEDORA: ['cron'],
            UBUNTU: ['cron'],
            DEBIAN: ['cron'],
        }
    
    def set_defaults(self):
        self.env.enabled = False
        self.env.cron_script_path = '/etc/cron.d/sshnice'
        self.env.cron_perms = '600'
    
    @task(precursors=['packager'])
    def configure(self):
        r = self.local_renderer
        if self.env.enabled:
            self.install_packages()
            remote_path = r.env.remote_path = self.env.cron_script_path
            r.put(
                local_path=self.find_template('sshnice/etc_crond_sshnice'),
                remote_path=remote_path, use_sudo=True)
            r.sudo('chown root:root %s' % remote_path)
            # Must be 600, otherwise gives INSECURE MODE error.
            # http://unix.stackexchange.com/questions/91202/cron-does-not-print-to-syslog
            r.sudo('chmod {cron_perms} {remote_path}')
            r.sudo('service cron restart')
        else:
            r.sudo('rm -f {cron_script_path}')
            r.sudo('service cron restart')

sshnice = SSHNiceSatchel()




from fabric.contrib.files import exists

from burlap.common import set_verbose
from burlap.selenium import selenium
from burlap.tests.functional_tests.base import TestCase
from burlap.deploy import thumbprint, clear_fs_cache, delete_plan_data_dir

class SeleniumTests(TestCase):

    def test_selenium(self):
        try:
            set_verbose(True)
            print('selenium.geckodriver_path:', selenium.geckodriver_path)
            selenium.genv.ROLE = 'local'
            selenium.genv.services = ['selenium']
            selenium.clear_caches()
    
            print('Enabling selenium/gecko to install and track old version.')
            selenium.env.enabled = True
            selenium.env.geckodriver_version = '0.13.0'
            selenium.clear_local_renderer()
            assert selenium.get_target_geckodriver_version_number() == '0.13.0'
            selenium.configure()
            clear_fs_cache()
            thumbprint(components=selenium.name)
            clear_fs_cache()
    
            # Confirm install succeeded.
            assert exists(selenium.geckodriver_path)
            assert not selenium.check_for_change()
            output = selenium.run('geckodriver --version')
            expected_version = selenium.env.geckodriver_version
            assert expected_version in output
            
            # Update configuration to track the most recent version.
            selenium.env.geckodriver_version = None
            selenium.clear_local_renderer()
            assert selenium.get_target_geckodriver_version_number() != '0.13.0'
            assert selenium.last_manifest.fingerprint == '0.13.0'
            
            print('Confirm we now see a pending change...')
            assert selenium.check_for_change()
            
            print('-'*80)
            print('Applying change...')
            selenium.configure()
            clear_fs_cache()
            delete_plan_data_dir()
            print('-'*80)
            print('Thumbprinting...')
            thumbprint(components=selenium.name)
            clear_fs_cache()
            print('-'*80)
            
            # Confirm the most recent version was installed.
            selenium.clear_caches()
            assert selenium.last_manifest.fingerprint == '0.15.0'
            output = selenium.run('geckodriver --version')
            expected_version = selenium.get_latest_geckodriver_version_number()
            assert expected_version in output
            assert not selenium.check_for_change()
    
            print('Update configuration to not manage gecko and apply...')
            selenium.env.enabled = False
            selenium.clear_local_renderer()
            selenium.configure()
    
            # Confirm gecko was uninstalled.
            assert not exists(selenium.geckodriver_path)

        finally:
            selenium.uninstall_geckodriver()



from __future__ import print_function

import time

from burlap.constants import *
from burlap import Satchel
from burlap.decorators import task

def get_boto():
    try:
        import boto
    except ImportError:
        boto = None
    return boto

class CloudfrontSatchel(Satchel):
    
    name = 'cloudfront'

    def set_defaults(self):
        pass

    @task
    def get_or_create_distribution(self, s3_bucket_name):
        assert isinstance(s3_bucket_name, basestring)
        boto = get_boto()
        origin_dns = '%s.s3.amazonaws.com' % s3_bucket_name
        if not self.dryrun:
            conn = boto.connect_cloudfront(
                self.genv.aws_access_key_id,
                self.genv.aws_secret_access_key
            )
            origin = boto.cloudfront.origin.S3Origin(origin_dns)
            
            distro = None
            dists = conn.get_all_distributions()
            for d in dists:
                print('Checking existing Cloudfront distribution %s...' % d.get_distribution().config.origin.dns_name)
                if origin_dns == d.get_distribution().config.origin.dns_name:
                    print('Found existing distribution!')
                    distro = d
                    break
                    
                # Necessary to avoid "Rate exceeded" errors.
                time.sleep(0.4)
            
            if not distro:
                print('Creating new distribution from %s...' % origin)
                distro = conn.create_distribution(origin=origin, enabled=True)
                
            return distro
        else:
            print('boto.connect_cloudfront().create_distribution(%s)' % repr(origin_dns))

    @task
    def configure(self):
        pass

cloudfront = CloudfrontSatchel()



"""
General tweaks and services to enhance system security.
"""
from __future__ import print_function

from burlap import Satchel
from burlap.constants import *
from burlap.decorators import task

class UnattendedUpgradesSatchel(Satchel):
    """
    Enables various degrees of automatic package download and installation.
    
    If you enable this satchel, it's highly encouraged you also enable and configure the postfix
    satchel, or some other way to receive email updates, so you'll be informed of upgrades and scheduled reboots.
    """
    
    name = 'unattendedupgrades'
    
    @property
    def packager_system_packages(self):
        return {
            UBUNTU: ['unattended-upgrades'],
            (UBUNTU, '12.04'): ['unattended-upgrades'],
            (UBUNTU, '14.04'): ['unattended-upgrades'],
            (UBUNTU, '16.04'): ['unattended-upgrades'],
            (UBUNTU, '16.14'): ['unattended-upgrades'],
            DEBIAN: ['unattended-upgrades'],
            RASPBIAN: ['unattended-upgrades'],
        }
    
    def set_defaults(self):
        
        self.env.mail_to = 'root@localhost'
        self.env.reboot = 'true'
        # This is relative to the system's timezone.
        # e.g. If system time is EST than 2:00 will be interpreted as 2am EST.
        self.env.reboot_time = '02:00'
        self.env.mailonlyonerror = "true"
        
        self.env.update_package_lists = 1
        self.env.download_upgradeable_packages = 1
        self.env.autoclean_interval = 7
        self.env.unattended_upgrade = 1

    @task(precursors=['packager'])
    def configure(self):
        
        os_version = self.os_version
        
        assert os_version.distro in (DEBIAN, RASPBIAN, UBUNTU), 'Unsupported OS: %s' % os_version.distro
        
        r = self.local_renderer
        
        if self.env.enabled:
            
            # Enable automatic package updates for Ubuntu.
            # Taken from the guide at https://help.ubuntu.com/lts/serverguide/automatic-updates.html.
            fn = self.render_to_file('unattendedupgrades/etc_apt_aptconfd_50unattended_upgrades')
            r.put(local_path=fn, remote_path='/etc/apt/apt.conf.d/50unattended-upgrades', use_sudo=True)
            fn = self.render_to_file('unattendedupgrades/etc_apt_aptconfd_10periodic')
            r.put(local_path=fn, remote_path='/etc/apt/apt.conf.d/10periodic', use_sudo=True)
            
        else:
            self.purge_packages()
        
UnattendedUpgradesSatchel()



from __future__ import print_function
import os
import sys
# import tempfile
import shutil # pylint: disable=unused-import
from commands import getstatusoutput

import yaml

from burlap.common import set_verbose, find_template, render_to_string
from burlap.tests.base import TestCase
from burlap.context import set_cwd

CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))

class ProjectTests(TestCase):
    
    def setUp(self):
        super(ProjectTests, self).setUp()
        
        set_verbose(True)
        # Ensure we're in burlap's root directory.
        os.chdir(os.path.abspath(os.path.join(CURRENT_DIR, '../..')))
    
    def test_project(self):
        try:
            project_dir = '/tmp/burlap_test_project'#tempfile.mkdtemp()
            if not os.path.isdir(project_dir):
                os.makedirs(project_dir)
            bin_dir = os.path.split(sys.executable)[0]
            
            with set_cwd(project_dir):
                cmd = (
                    '. {bin_dir}/activate; '
                    'burlap-admin.py skel myproject'
                ).format(**locals())
                print(cmd)
                ret = os.system(cmd)
                print('ret:', ret)
                assert not ret
                
                cmd = (
                    '. {bin_dir}/activate; '
                    'burlap-admin.py add-role prod dev'
                ).format(**locals())
                print(cmd)
                ret = os.system(cmd)
                print('ret:', ret)
                assert not ret

                if not os.path.isdir('satchels'):
                    os.makedirs('satchels')
                os.system('touch satchels/__init__.py')
                open('satchels/junk.py', 'w').write("""
from burlap import Satchel
from burlap.constants import *
from burlap.decorators import task

class JunkSatchel(Satchel):

    name = 'junk'

    def set_defaults(self):
        self.env.param = 'default'

    @task
    def show_param(self):
        print('param:%s' % self.env.param)

    @task
    def configure(self):
        self.show_param()
        
junk = JunkSatchel()
""")
                
                open('roles/all/settings.yaml', 'w').write(yaml.dump(dict(
                    app_name='myproject_site',
                    default_site='myproject',
                    services=['junk'],
                    sites={},
                    junk_param='allvalue',
                )))
                
                open('roles/prod/settings.yaml', 'w').write(yaml.dump(dict(
                    inherits='all',
                    hosts=['localhost'],
                    junk_enabled=True,
                    junk_param='prodvalue',
                )))
                
                open('roles/dev/settings.yaml', 'w').write(yaml.dump(dict(
                    inherits='all',
                    hosts=['localhost'],
                    junk_enabled=True,
                    junk_param='devvalue',
                )))

                ## Check prod role.
                
                os.system('rm -Rf .burlap')
                cmd = (
                    '. {bin_dir}/activate; '
                    'fab prod junk.show_param'
                ).format(**locals())
                print(cmd)
                status, output = getstatusoutput(cmd)
                print('output:', output)
                assert 'param:prodvalue' in output

                os.system('rm -Rf .burlap')
                cmd = (
                    '. {bin_dir}/activate; '
                    'fab prod deploy.preview:verbose=1'
                ).format(**locals())
                print(cmd)
                status, output = getstatusoutput(cmd)
                print('output:', output)
                assert 'junk.configure' in output

                os.system('rm -Rf .burlap')
                cmd = (
                    '. {bin_dir}/activate; '
                    'fab prod deploy.run:yes=1'
                ).format(**locals())
                print(cmd)
                status, output = getstatusoutput(cmd)
                print('output:', output)
                assert 'param:prodvalue' in output

                ## Check dev role.

                os.system('rm -Rf .burlap')
                cmd = (
                    '. {bin_dir}/activate; '
                    'fab dev junk.show_param'
                ).format(**locals())
                print(cmd)
                status, output = getstatusoutput(cmd)
                print('output:', output)
                assert 'param:devvalue' in output
                
                os.system('rm -Rf .burlap')
                cmd = (
                    '. {bin_dir}/activate; '
                    'fab dev deploy.preview'
                ).format(**locals())
                print(cmd)
                status, output = getstatusoutput(cmd)
                print('output:', output)
                assert 'junk.configure' in output

                os.system('rm -Rf .burlap')
                cmd = (
                    '. {bin_dir}/activate; '
                    'fab dev deploy.run:yes=1'
                ).format(**locals())
                print(cmd)
                status, output = getstatusoutput(cmd)
                print('output:', output)
                assert 'param:devvalue' in output

        finally:
            #shutil.rmtree(project_dir)
            pass
    
    
    def test_find_template(self):
        fn = 'burlap/gitignore.template'
        ret = find_template(fn)
        print('ret:', ret)
        assert ret and ret.endswith(fn) 
    
    
    def test_render_to_string(self):
        ret = render_to_string(
            'postfix/etc_postfix_sasl_sasl_passwd',
            dict(
                postfix_host='smtp.test.com',
                postfix_port=1234,
                postfix_username='myusername',
                postfix_password='mypassword',
            ))
        print('ret:', ret)
        assert ret == "[smtp.test.com]:1234 myusername:mypassword"



from mock import patch

from burlap.constants import *
from burlap.common import get_satchel, set_verbose, set_dryrun, OS
from burlap.tests.base import TestCase

class ServiceTests(TestCase):
    
    def test_post_deploy(self):
        with patch('burlap.common.get_os_version') as mock_get_os_version:
            mock_get_os_version.return_value = OS(
                type=LINUX,
                distro=UBUNTU,
                release='16.04')
            
            service = get_satchel('service')
            apache = get_satchel('apache')
            service.genv.services.append(apache.name)
            set_verbose(1)
            set_dryrun(1)
            print('self.genv.services:', service.genv.services)
            service.post_deploy()



"""
Supervisor processes
====================

This module provides high-level tools for managing long-running
processes using `supervisord`_.

.. _supervisord: http://supervisord.org/

"""
from __future__ import print_function

import os
import time

#from fabric.api import hide, settings

from burlap.constants import *
from burlap import ServiceSatchel
#from burlap.utils import run_as_root
from burlap.decorators import task

# def reload_config():
#     """
#     Reload supervisor configuration.
#     """
#     run_as_root("supervisorctl reload")
# 
# 
# def update_config():
#     """
#     Reread and update supervisor job configurations.
# 
#     Less heavy-handed than a full reload, as it doesn't restart the
#     backend supervisor process and all managed processes.
#     """
#     run_as_root("supervisorctl update")
# 
# 
# def process_status(name):
#     """
#     Get the status of a supervisor process.
#     """
#     with settings(hide('running', 'stdout', 'stderr', 'warnings'), warn_only=True):
#         res = run_as_root("supervisorctl status {name}" % locals())
#         if res.startswith("No such process"):
#             return None
#         else:
#             return res.split()[1]
# 
# 
# def start_process(name):
#     """
#     Start a supervisor process
#     """
#     run_as_root("supervisorctl start {name}" % locals())
# 
# 
# def stop_process(name):
#     """
#     Stop a supervisor process
#     """
#     run_as_root("supervisorctl stop {name}" % locals())
# 
# 
# def restart_process(name):
#     """
#     Restart a supervisor process
#     """
#     run_as_root("supervisorctl restart {name}" % locals())


class SupervisorSatchel(ServiceSatchel):
    
    name = 'supervisor'
    
    ## Service options.
    
    #ignore_errors = True
    
    post_deploy_command = 'restart'
    
    @property
    def packager_system_packages(self):
        return {
            UBUNTU: ['supervisor'],
        }
    
    def set_defaults(self):
    
        self.env.config_template = 'supervisor/supervisor_daemon.template2.config'
        self.env.config_path = '/etc/supervisor/supervisord.conf'
        #/etc/supervisor/conf.d/celery_
        self.env.conf_dir = '/etc/supervisor/conf.d'
        self.env.daemon_bin_path_template = '{pip_virtualenv_dir}/bin/supervisord'
        self.env.daemon_path = '/etc/init.d/supervisord'
        self.env.bin_path_template = '{pip_virtualenv_dir}/bin'
        self.env.daemon_pid = '/var/run/supervisord.pid'
        self.env.log_path = "/var/log/supervisord.log"
        self.env.supervisorctl_path_template = '{pip_virtualenv_dir}/bin/supervisorctl'
        self.env.kill_pattern = ''
        self.env.max_restart_wait_minutes = 5
        
        self.env.services = []
        
        # Functions that, when called, should return a supervisor service text
        # ready to be appended to supervisord.conf.
        # It will be called once for each site.
        self.genv._supervisor_create_service_callbacks = []
        
        self.env.service_commands = {
            START:{
                FEDORA: 'systemctl start supervisord.service',
                UBUNTU: 'service supervisor start',
            },
            STOP:{
                FEDORA: 'systemctl stop supervisor.service',
                UBUNTU: 'service supervisor stop',
            },
            DISABLE:{
                FEDORA: 'systemctl disable httpd.service',
                UBUNTU: 'chkconfig supervisor off',
            },
            ENABLE:{
                FEDORA: 'systemctl enable httpd.service',
                UBUNTU: 'chkconfig supervisor on',
            },
            RESTART:{
                FEDORA: 'systemctl restart supervisord.service',
                UBUNTU: 'service supervisor restart; sleep 5',
            },
            STATUS:{
                FEDORA: 'systemctl --no-pager status supervisord.service',
                (UBUNTU, '14.04'): 'service supervisor status',
                (UBUNTU, '16.04'): 'systemctl --no-pager status supervisor',
                UBUNTU: 'systemctl --no-pager status supervisor',
            },
        }
    
    def render_paths(self):
        r = self.local_renderer
        r.env.daemon_bin_path = r.format(r.env.daemon_bin_path_template)
        r.env.bin_path = r.format(r.env.bin_path_template)
        r.env.supervisorctl_path = r.format(r.env.supervisorctl_path_template)
    
    def register_callback(self, f):
        self.genv._supervisor_create_service_callbacks.append(f)
        
    def restart(self):
        """
        Supervisor can take a very long time to start and stop,
        so wait for it.
        """
        n = 60
        sleep_n = int(self.env.max_restart_wait_minutes/10.*60)
        for _ in xrange(n):
            self.stop()
            if self.dryrun or not self.is_running():
                break
            print('Waiting for supervisor to stop (%i of %i)...' % (_, n))
            time.sleep(sleep_n)
        self.start()
        for _ in xrange(n):
            if self.dryrun or self.is_running():
                return
            print('Waiting for supervisor to start (%i of %i)...' % (_, n))
            time.sleep(sleep_n)
        raise Exception('Failed to restart service %s!' % self.name)
    
    def record_manifest(self):
        """
        Called after a deployment to record any data necessary to detect changes
        for a future deployment.
        """
        data = super(SupervisorSatchel, self).record_manifest()
        
        # Celery deploys itself through supervisor, so monitor its changes too in Apache site configs.
        for site_name, site_data in self.genv.sites.iteritems():
            if self.verbose:
                print(site_name, site_data)
            data['celery_has_worker_%s' % site_name] = site_data.get('celery_has_worker', False)
        
        data['configured'] = True
        
        return data

    @task
    def write_configs(self, site=None):
        
        verbose = self.verbose
        
        self.render_paths()
        
        supervisor_services = []
        process_groups = []
        
        #TODO:check available_sites_by_host and remove dead?
        for site, site_data in self.iter_sites(site=site, renderer=self.render_paths):
            if verbose:
                print('write_configs.site:', site)
            for cb in self.genv._supervisor_create_service_callbacks:
                ret = cb(site=site)
                if isinstance(ret, basestring):
                    supervisor_services.append(ret)
                elif isinstance(ret, tuple):
                    assert len(ret) == 2
                    conf_name, conf_content = ret
                    if verbose:
                        print('conf_name:', conf_name)
                        print('conf_content:', conf_content)
                    remote_fn = os.path.join(self.env.conf_dir, conf_name)
                    local_fn = self.write_to_file(conf_content)
                    self.put_or_dryrun(local_path=local_fn, remote_path=remote_fn, use_sudo=True)
                    
                    process_groups.append(os.path.splitext(conf_name)[0])
                    
        self.env.services_rendered = '\n'.join(supervisor_services)
    
        fn = self.render_to_file(self.env.config_template)
        self.put_or_dryrun(local_path=fn, remote_path=self.env.config_path, use_sudo=True)

    def deploy_services(self, site=None):
        """
        Collects the configurations for all registered services and writes
        the appropriate supervisord.conf file.
        """
        
        verbose = self.verbose
        
        r = self.local_renderer
#         
#         target_sites = self.genv.available_sites_by_host.get(hostname, None)
        
        self.render_paths()
        
        supervisor_services = []
        
        process_groups = []
        
        r.sudo('rm -Rf /etc/supervisor/conf.d/*')
        
        #TODO:check available_sites_by_host and remove dead?
        self.write_configs(site=site)
        for site, site_data in self.iter_sites(site=site, renderer=self.render_paths):
            if verbose:
                print('deploy_services.site:', site)
                
            # Only load site configurations that are allowed for this host.
#             if target_sites is not None:
#                 assert isinstance(target_sites, (tuple, list))
#                 if site not in target_sites:
#                     continue
                
            for cb in self.genv._supervisor_create_service_callbacks:
                if self.verbose:
                    print('cb:', cb)
                ret = cb(site=site)
                if self.verbose:
                    print('ret:', ret)
                if isinstance(ret, basestring):
                    supervisor_services.append(ret)
                elif isinstance(ret, tuple):
                    assert len(ret) == 2
                    conf_name, conf_content = ret
#                     print('conf_name:', conf_name)
#                     print('conf_content:', conf_content)
                    remote_fn = os.path.join(self.env.conf_dir, conf_name)
                    local_fn = self.write_to_file(conf_content)
#                     
                    process_groups.append(os.path.splitext(conf_name)[0])
                    
        self.env.services_rendered = '\n'.join(supervisor_services)
    
        fn = self.render_to_file(self.env.config_template)
        r.put(local_path=fn, remote_path=self.env.config_path, use_sudo=True)
        
        # We use supervisorctl to configure supervisor, but this will throw a uselessly vague
        # error message is supervisor isn't running.
        if not self.is_running():
            self.start()
            
        for pg in process_groups:
            r.env.pg = pg
            r.sudo('supervisorctl add {pg}')
        
        #TODO:are all these really necessary?
        r.sudo('supervisorctl reread')
        #r.sudo('supervisorctl update')
        r.sudo('supervisorctl restart all')
    
    @task(precursors=['packager', 'user', 'rabbitmq'])
    def configure(self, **kwargs):
        kwargs['site'] = ALL
        
#         last_manifest = supervisor_satchel.last_manifest
#         if not last_manifest or not last_manifest.get('configured'):
#             configure()
        
        self.deploy_services(**kwargs)
        
supervisor = SupervisorSatchel()



 
from burlap.common import set_dryrun, run_or_dryrun
from burlap.files import file # pylint: disable=redefined-builtin
from burlap.tests.functional_tests.base import TestCase

is_file = file.is_file

class CommonTests(TestCase):
    
    def test_dryrun(self):
        
        set_dryrun(True)
        run_or_dryrun('touch ~/abc.txt')
        assert not is_file('~/abc.txt')
        
        set_dryrun(1)
        run_or_dryrun('touch ~/def.txt')
        assert not is_file('~/def.txt')
        
        set_dryrun(False)
        run_or_dryrun('touch ~/mno.txt')
        assert not is_file('~/mno.txt')
        run_or_dryrun('rm -f ~/mno.txt')
        
        set_dryrun(0)
        run_or_dryrun('touch ~/xyz.txt')
        assert not is_file('~/xyz.txt')
        run_or_dryrun('rm -f ~/xyz.txt')



from __future__ import print_function

from burlap import ServiceSatchel
from burlap.constants import *
from burlap.decorators import task

class SnortSatchel(ServiceSatchel):
    
    name = 'snort'
    
    def set_defaults(self):
        super(SnortSatchel, self).set_defaults()

    @property
    def packager_system_packages(self):
        return {
            DEBIAN: [
                'snort',
            ],
            UBUNTU: [
                'snort',
            ],
        }
    
    @task(precursors=['packager'])
    def configure(self):
        pass

snort = SnortSatchel()



PACKAGERS = APT, YUM = ('apt-get', 'yum')

APT_KEY = 'apt-key'
APT_SOURCE = 'apt-source'

OS_TYPES = LINUX, WINDOWS = ('linux', 'windows')
OS_DISTRO = (
    FEDORA,
    UBUNTU,
    DEBIAN,
    RASPBIAN,
    CENTOS,
    SUNOS,
    ARCH,
    REDHAT,
    SLES,
    GENTOO,
) = (
    'fedora',
    'ubuntu',
    'debian',
    'raspbian',
    'centos',
    'sunos',
    'arch',
    'redhat',
    'sles',
    'gentoo',
)

SUN = 'sun'

WHEEZY = 'wheezy'
JESSIE = 'jessie'

SYSTEM = 'system'
RUBY = 'ruby'
PYTHON = 'python'
PACKAGE_TYPES = (
    SYSTEM,
    PYTHON, # pip
    RUBY, # gem
)

ALL = 'all' # denotes the global role

START = 'start'
STOP = 'stop'
STATUS = 'status'
RELOAD = 'reload'
RESTART = 'restart'
ENABLE = 'enable'
DISABLE = 'disable'
STATUS = 'status'
SERVICE_COMMANDS = (
    START,
    STOP,
    STATUS,
    RESTART,
    ENABLE,
    DISABLE,
    STATUS,
)

DJANGO = 'DJANGO'

SITE = 'SITE'
ROLE = 'ROLE'

LOCALHOSTS = ('localhost', '127.0.0.1')

UTF8 = 'UTF8'



"""
Various debug tasks.

Note, this is a special module, in that all tasks defined here are auto-imported
into the top-level namespace. That means you access them by calling them directly,
not through "debug."
"""
from __future__ import print_function

import re
from pprint import pprint

from burlap import ContainerSatchel
from burlap.decorators import task, runs_once

def list_to_str_or_unknown(lst):
    if len(lst):
        return ', '.join(map(str, lst))
    return 'unknown'

class DebugSatchel(ContainerSatchel):
    
    name = 'debug'
    
    def set_defaults(self):
        self.env.shell_default_dir = '~'
        self.env.shell_interactive_cmd = '/bin/bash -i'
        self.env.shell_default_options = ['-o StrictHostKeyChecking=no']
    
    @task
    def ping_servers(self):
        self.local('nmap -p 80 -sT {host_string}')
    
    @task
    def list_settings(self, name):
        from burlap import load_yaml_settings
        load_yaml_settings(name=name, verbose=1)
    
    @task
    def test_dryrun1(self):
        print('test1.get_dryrun:', self.dryrun)#should show false
        self.local('echo "hello 1"')
        self.test_dryrun2(dryrun=1)
        self.local('echo "hello 3"')
        
    @task
    def test_dryrun2(self):
        print('test2.get_dryrun:', self.dryrun)#should show true
        self.local('echo "hello 2"')
    
    @task
    def list_env(self, key=None):
        """
        Displays a list of environment key/value pairs.
        """
        for k, v in sorted(self.genv.iteritems(), key=lambda o: o[0]):
            if key and k != key:
                continue
            print('%s ' % (k,))
            pprint(v, indent=4)

    @task
    def list_sites(self, site='all', *args, **kwargs):
        kwargs['site'] = site
        for site, data in self.iter_sites(*args, **kwargs):
            print(site)

    @task
    def list_server_specs(self, cpu=1, memory=1, hdd=1):
        """
        Displays a list of common servers characteristics, like number
        of CPU cores, amount of memory and hard drive capacity.
        """
        
        cpu = int(cpu)
        memory = int(memory)
        hdd = int(hdd)
        
        # CPU
        if cpu:
            cmd = 'cat /proc/cpuinfo | grep -i "model name"'
            ret = self.run(cmd)
            matches = map(str.strip, re.findall(r'model name\s+:\s*([^\n]+)', ret, re.DOTALL|re.I))
            cores = {}
            for match in matches:
                cores.setdefault(match, 0)
                cores[match] += 1
        
        # Memory
        if memory:
            cmd = 'dmidecode --type 17'
            ret = self.sudo(cmd)
            #print repr(ret)
            matches = re.findall(r'Memory\s+Device\r\n(.*?)(?:\r\n\r\n|$)', ret, flags=re.DOTALL|re.I)
            #print len(matches)
            #print matches[0]
            memory_slot_dicts = []
            for match in matches:
                attrs = dict([(_a.strip(), _b.strip()) for _a, _b in re.findall(r'^([^:]+):\s+(.*)$', match, flags=re.MULTILINE)])
                #print attrs
                memory_slot_dicts.append(attrs)
            total_memory_gb = 0
            total_slots_filled = 0
            total_slots = len(memory_slot_dicts)
            memory_types = set()
            memory_forms = set()
            memory_speeds = set()
            for memory_dict in memory_slot_dicts:
                try:
                    size = int(round(float(re.findall(r'([0-9]+)\s+MB', memory_dict['Size'])[0])/1024.))
                    #print size
                    total_memory_gb += size
                    total_slots_filled += 1
                except IndexError:
                    pass
                _v = memory_dict['Type']
                if _v != 'Unknown':
                    memory_types.add(_v)
                _v = memory_dict['Form Factor']
                if _v != 'Unknown':
                    memory_forms.add(_v)
                _v = memory_dict['Speed']
                if _v != 'Unknown':
                    memory_speeds.add(_v)
        
        # Storage
        if hdd:
            #cmd = 'ls /dev/*d* | grep "/dev/[a-z]+d[a-z]$"'
            cmd = 'find /dev -maxdepth 1 | grep -E "/dev/[a-z]+d[a-z]$"'
            devices = map(str.strip, self.run(cmd).split('\n'))
            total_drives = len(devices)
            total_physical_storage_gb = 0
            total_logical_storage_gb = 0
            drive_transports = set()
            for device in devices:
                cmd = 'udisks --show-info %s |grep -i "  size:"' % (device)
                ret = self.run(cmd)
                size_bytes = float(re.findall(r'size:\s*([0-9]+)', ret)[0].strip())
                size_gb = int(round(size_bytes/1024/1024/1024))
                #print device, size_gb
                total_physical_storage_gb += size_gb
                
                with self.settings(warn_only=True):
                    cmd = 'hdparm -I %s|grep -i "Transport:"' % device
                    ret = self.sudo(cmd)
                    if ret and not ret.return_code:
                        drive_transports.add(ret.split('Transport:')[-1].strip())
                    
            cmd = "df | grep '^/dev/[mhs]d*' | awk '{s+=$2} END {print s/1048576}'"
            ret = self.run(cmd)
            total_logical_storage_gb = float(ret)
        
        if cpu:
            print('-'*80)
            print('CPU')
            print('-'*80)
            type_str = ', '.join(['%s x %i' % (_type, _count) for _type, _count in cores.items()])
            print('Cores: %i' % sum(cores.values()))
            print('Types: %s' % type_str)
        
        if memory:
            print('-'*80)
            print('MEMORY')
            print('-'*80)
            print('Total: %s GB' % total_memory_gb)
            print('Type: %s' % list_to_str_or_unknown(memory_types))
            print('Form: %s' % list_to_str_or_unknown(memory_forms))
            print('Speed: %s' % list_to_str_or_unknown(memory_speeds))
            print('Slots: %i (%i filled, %i empty)' % (total_slots, total_slots_filled, total_slots - total_slots_filled))
        
        if hdd:
            print('-'*80)
            print('STORAGE')
            print('-'*80)
            print('Total physical drives: %i' % total_drives)
            print('Total physical storage: %s GB' % total_physical_storage_gb)
            print('Total logical storage: %s GB' % total_logical_storage_gb)
            print('Types: %s' % list_to_str_or_unknown(drive_transports))

    @task
    def list_hosts(self):
        print('hosts:', self.genv.hosts)
    
    
    @task
    def info(self):
        print('Info')
        print('\tROLE:', self.genv.ROLE)
        print('\tSITE:', self.genv.SITE)
        print('\tdefault_site:', self.genv.default_site)
    
    
    @task
    @runs_once
    def shell(self, gui=0, command=''):
        """
        Opens an SSH connection.
        """
        from burlap.common import get_hosts_for_site
        r = self.local_renderer
        
        if r.genv.SITE != r.genv.default_site:
            shell_hosts = get_hosts_for_site()
            if shell_hosts:
                r.genv.host_string = shell_hosts[0]
        
        r.env.SITE = r.genv.SITE or r.genv.default_site
        
        if int(gui):
            r.env.shell_default_options.append('-X')
        
        if 'host_string' not in self.genv or not self.genv.host_string:
            if 'available_sites' in self.genv and r.env.SITE not in r.genv.available_sites:
                raise Exception('No host_string set. Unknown site %s.' % r.env.SITE)
            else:
                raise Exception('No host_string set.')
        
        if '@' in r.genv.host_string:
            r.env.shell_host_string = r.genv.host_string
        else:
            r.env.shell_host_string = '{user}@{host_string}'
            
        if command:
            r.env.shell_interactive_cmd_str = command
        else:
            r.env.shell_interactive_cmd_str = r.format(r.env.shell_interactive_cmd)
        
        r.env.shell_default_options_str = ' '.join(r.env.shell_default_options)
        if self.is_local:
            self.vprint('Using direct local.')
            cmd = '{shell_interactive_cmd_str}'
        elif r.genv.key_filename:
            self.vprint('Using key filename.')
            # If host_string contains the port, then strip it off and pass separately.
            port = r.env.shell_host_string.split(':')[-1]
            if port.isdigit():
                r.env.shell_host_string = r.env.shell_host_string.split(':')[0] + (' -p %s' % port)
            cmd = 'ssh -t {shell_default_options_str} -i {key_filename} {shell_host_string} "{shell_interactive_cmd_str}"'
        elif r.genv.password:
            self.vprint('Using password.')
            cmd = 'ssh -t {shell_default_options_str} {shell_host_string} "{shell_interactive_cmd_str}"'
        else:
            # No explicit password or key file needed?
            self.vprint('Using nothing.')
            cmd = 'ssh -t {shell_default_options_str} {shell_host_string} "{shell_interactive_cmd_str}"'
        r.local(cmd)

    @task
    def run(self, command):
        with self.settings(warn_only=True):
            self.run(command)

    @task
    def disk(self):
        """
        Display percent of disk usage.
        """
        r = self.local_renderer
        r.run(r.env.disk_usage_command)

    @task
    def tunnel(self, local_port, remote_port):
        """
        Creates an SSH tunnel.
        """
        r = self.local_renderer
        r.env.tunnel_local_port = local_port
        r.env.tunnel_remote_port = remote_port
        r.local(' ssh -i {key_filename} -L {tunnel_local_port}:localhost:{tunnel_remote_port} {user}@{host_string} -N')
    
    @task
    def test_local(self):
        self.local("echo hello")
    
    @task
    def test_run(self):
        self.run("echo hello")
    
    @task
    def test_sudo(self):
        self.sudo("echo hello")

debug = DebugSatchel()



from __future__ import print_function

import sys
import traceback

from burlap import Satchel
from burlap.constants import *
from burlap.decorators import task

# from fabric.api import (
#     env,
#     settings,
#     hide,
# )


# from burlap import systemd
# from burlap.system import using_systemd, distrib_family
# from burlap.utils import run_as_root

from burlap import common
# from burlap.common import (
#     get_dryrun,
# )
# from burlap.decorators import task_or_dryrun

#run_as_root = sudo_or_dryrun

# @task
# def is_running(service):
#     """
#     Check if a service is running.
# 
#     ::
# 
#         import burlap
# 
#         if burlap.service.is_running('foo'):
#             print("Service foo is running!")
#     """
#     
#     #DEPRECATED?
#     service = service.strip().lower()
#     _ran = False
#     for _service in self.genv.services:
#         _service = _service.strip().upper()
#         if service and _service.lower() != service:
#             continue
#         srv = common.services.get(_service)
#         if srv:
#             _ran = True
#             #print('%s.is_running: %s' % (service, srv.is_running()))
#             return srv.is_running()
#             
#     with settings(hide('running', 'stdout', 'stderr', 'warnings'),
#                   warn_only=True):
#         if using_systemd():
#             return systemd.is_running(service)
#         else:
#             if distrib_family() != "gentoo":
#                 test_upstart = run_as_root('test -f /etc/init/%s.conf' %
#                                            service)
#                 status = _run_service(service, 'status')
#                 if test_upstart.succeeded:
#                     return 'running' in status
#                 else:
#                     return status.succeeded
#             else:
#                 # gentoo
#                 status = _run_service(service, 'status')
#                 return ' started' in status
# 
# 
# def start(service):
#     """
#     Start a service.
# 
#     ::
# 
#         import burlap
# 
#         # Start service if it is not running
#         if not burlap.service.is_running('foo'):
#             burlap.service.start('foo')
#     """
#     
#     with settings(warn_only=True):
#         _run_service(service, 'start')
#         
#     # Sometimes race conditions result in us trying to start a service that wasn't running
#     # when we checked, but started running by the time we tried to start.
#     # So ignore the error that the service we're trying to start has started
#     # and then do a separate check to confirm the service has started.
#     assert is_running(service), 'Service %s failed to start.' % service
#     
# 
# @task
# def stop(service=''):
#     """
#     Stop a service.
# 
#     ::
# 
#         import burlap
# 
#         # Stop service if it is running
#         if burlap.service.is_running('foo'):
#             burlap.service.stop('foo')
#     """
#     
#     ran = False
#     service = service.strip().lower()
#     for _service in self.genv.services:
#         _service = _service.strip().upper()
#         if service and _service.lower() != service:
#             continue
#         funcs = common.service_stoppers.get(_service)
#         if funcs:
#             print('Restarting service %s...' % (_service,))
#             for func in funcs:
#                 func()
#                 ran = True
#                 
#     if not ran and not get_dryrun() and service:
#         _run_service(service, 'stop')
# 
# 
# @task
# def restart(service=''):
#     """
#     Restart a service.
# 
#     ::
# 
#         import burlap
# 
#         # Start service, or restart it if it is already running
#         if burlap.service.is_running('foo'):
#             burlap.service.restart('foo')
#         else:
#             burlap.service.start('foo')
#     """
#     
#     service = service.strip().lower()
#     _ran = False
# 
#     for _service in self.genv.services:
#         _service = _service.strip().upper()
# 
#         if service and _service.lower() != service:
#             continue
#             
#         srv = common.services.get(_service)
#         if srv:
#             srv.restart()
#             _ran = True
#             continue
#             
#         funcs = common.service_restarters.get(_service)
#         if funcs:
#             print('Restarting service %s...' % (_service,))
#             for func in funcs:
#                 if not self.dryrun:
#                     func()
#                     _ran = True
#     
#     if not get_dryrun() and not _ran and service:
#         _run_service(service, 'restart')
# 
# 
# def reload(service): # pylint: disable=redefined-builtin
#     """
#     Reload a service.
# 
#     ::
# 
#         import burlap
# 
#         # Reload service
#         burlap.service.reload('foo')
# 
#     .. warning::
# 
#         The service needs to support the ``reload`` operation.
#     """
#     _run_service(service, 'reload')
# 
# 
# def force_reload(service):
#     """
#     Force reload a service.
# 
#     ::
# 
#         import burlap
# 
#         # Force reload service
#         burlap.service.force_reload('foo')
# 
#     .. warning::
# 
#         The service needs to support the ``force-reload`` operation.
#     """
#     _run_service(service, 'force-reload')
# 
# 
# def _run_service(service, action):
#     """
#     Compatibility layer for distros that use ``service`` and those that don't.
#     """
#     if distrib_family() != "gentoo":
#         status = run_as_root('service %(service)s %(action)s' % locals(),
#                              pty=False)
#     else:
#         # gentoo
#         status = run_as_root('/etc/init.d/%(service)s %(action)s' % locals(),
#                              pty=False)
#     return status

class ServiceManagementSatchel(Satchel):
    
    name = 'service'
    
    @task
    def pre_deploy(self):
        """
        Runs methods services have requested be run before each deployment.
        """
        for service in self.genv.services:
            service = service.strip().upper()
            funcs = common.service_pre_deployers.get(service)
            if funcs:
                print('Running pre-deployments for service %s...' % (service,))
                for func in funcs:
                    func()
                    
    @task
    def deploy(self):
        """
        Applies routine, typically application-level changes to the service.
        """
        for service in self.genv.services:
            service = service.strip().upper()
            funcs = common.service_deployers.get(service)
            if funcs:
                print('Deploying service %s...' % (service,))
                for func in funcs:
                    if not self.dryrun:
                        func()
    
    @task
    def post_deploy(self):
        """
        Runs methods services have requested be run before after deployment.
        """
        for service in self.genv.services:
            service = service.strip().upper()
            self.vprint('post_deploy:', service)
            funcs = common.service_post_deployers.get(service)
            if funcs:
                self.vprint('Running post-deployments for service %s...' % (service,))
                for func in funcs:
                    try:
                        func()
                    except Exception as e:
                        print('Post deployment error: %s' % e, file=sys.stderr) 
                        print(traceback.format_exc(), file=sys.stderr)
    
    def is_selected(self, name):
        name = name.strip().upper()
        for service in self.genv.services:
            if service.strip().upper() == name:
                return True
        return False
    
    @task
    def pre_db_dump(self):
        """
        Runs methods services that have requested to be run before each
        database dump.
        """
        for service in self.genv.services:
            service = service.strip().upper()
            funcs = common.service_pre_db_dumpers.get(service)
            if funcs:
                print('Running pre-database dump for service %s...' % (service,))
                for func in funcs:
                    func()
    
    @task
    def post_db_dump(self):
        """
        Runs methods services that have requested to be run before each
        database dump.
        """
        for service in self.genv.services:
            service = service.strip().upper()
            funcs = common.service_post_db_dumpers.get(service)
            if funcs:
                print('Running post-database dump for service %s...' % (service,))
                for func in funcs:
                    func()


    @task
    def refresh(self):
        self.configure()
        self.deploy()
        self.post_deploy()

    @task
    def configure(self):
        """
        Applies one-time settings changes to the host, usually to initialize the service.
        """
        print('env.services:', self.genv.services)
        for service in list(self.genv.services):
            service = service.strip().upper()
            funcs = common.service_configurators.get(service, [])
            if funcs:
                print('!'*80)
                print('Configuring service %s...' % (service,))
                for func in funcs:
                    print('Function:', func)
                    if not self.dryrun:
                        func()

service = ServiceManagementSatchel()



from __future__ import with_statement

import os
import contextlib

#http://stackoverflow.com/a/169112/247542
@contextlib.contextmanager
def set_cwd(new_path):
    """
    Usage:
    
        with set_cwd('/some/dir'):
            walk_around_the_filesystem()
    """
    try:
        curdir = os.getcwd()
    except OSError:
        curdir = new_path
    try:
        os.chdir(new_path)
        yield
    finally:
        os.chdir(curdir)



"""
Inspects and manipulates settings files.
"""
from __future__ import print_function

from pprint import pprint
import types

from fabric.api import (
    env,
)

from burlap.decorators import task_or_dryrun


@task_or_dryrun
def show(keyword=''):
    """
    Displays a list of all environment key/value pairs for the current role.
    """
    keyword = keyword.strip().lower()
    max_len = max(len(k) for k in env.iterkeys())
    keyword_found = False
    for k in sorted(env.iterkeys()):
        if keyword and keyword not in k.lower():
            continue
        keyword_found = True
        #print '%s: %s' % (k, env[k])
        print('%s: ' % (k.ljust(max_len),))
        pprint(env[k], indent=4)
    if keyword:
        if not keyword_found:
            print('Keyword "%s" not found.' % keyword)


@task_or_dryrun
def record_manifest():
    """
    Called after a deployment to record any data necessary to detect changes
    for a future deployment.
    """
    data = {}
    # Record settings.
    data['settings'] = dict(
        (k, v)
        for k, v in env.iteritems()
        if not isinstance(v, types.GeneratorType) and k.strip() and not k.startswith('_') and not callable(v)
    )
    # Record tarball hash.
    # Record database migrations.
    # Record media hash.
    return data


def compare_manifest(data=None):
    """
    Called before a deployment, given the data returned by record_manifest(),
    for determining what, if any, tasks need to be run to make the target
    server reflect the current settings within the current context.
    """



from __future__ import absolute_import, print_function

from pprint import pprint

from burlap.common import Satchel
from burlap.constants import *
from burlap.decorators import task

EZ_SETUP = 'ez_setup'
PYTHON_PIP = 'python-pip'
GET_PIP = 'get-pip'
BOOTSTRAP_METHODS = (
    EZ_SETUP,
    PYTHON_PIP,
    GET_PIP,
)

class PIPSatchel(Satchel):
    
    name = 'pip'
    
    @property
    def packager_system_packages(self):
        return {
            UBUNTU: [
                'gcc', 'python-dev', 'build-essential', 'python-pip',
            ],
            (UBUNTU, '14.04'): [
                #'python-pip',#obsolete in 14.04?
                #'python-virtualenv',#obsolete in 14.04?
                'gcc', 'python-dev', 'build-essential', 'python-pip',
            ],
            (UBUNTU, '14.10'): [
                'gcc', 'python-dev', 'build-essential', 'python-pip',
            ],
            (UBUNTU, '16.04'): [
                'gcc', 'python-dev', 'build-essential', 'python-pip',
            ],
            (UBUNTU, '16.10'): [
                'gcc', 'python-dev', 'build-essential', 'python-pip',
            ],
        }
        
    def set_defaults(self):
        self.env.bootstrap_method = GET_PIP
        self.env.check_permissions = True
        self.env.user = 'www-data'
        self.env.group = 'www-data'
        self.env.perms = '775'
        self.env.virtualenv_dir = '.env'
        self.env.requirements = 'pip-requirements.txt'

    @task
    def has_pip(self):
        r = self.local_renderer
        with self.settings(warn_only=True):
            ret = (r.run('which pip') or '').strip()
            return bool(ret)
            
    @task
    def bootstrap(self, force=0):
        """
        Installs all the necessary packages necessary for managing virtual
        environments with pip.
        """
        force = int(force)
        if self.has_pip() and not force:
            return
        
        r = self.local_renderer
        
        if r.env.bootstrap_method == GET_PIP:
            r.sudo('curl --silent --show-error --retry 5 https://bootstrap.pypa.io/get-pip.py | python')
        elif r.env.bootstrap_method == EZ_SETUP:
            r.run('wget http://peak.telecommunity.com/dist/ez_setup.py -O /tmp/ez_setup.py')
            with self.settings(warn_only=True):
                r.sudo('python /tmp/ez_setup.py -U setuptools')
            r.sudo('easy_install -U pip')
        elif r.env.bootstrap_method == PYTHON_PIP:
            r.sudo('apt-get install -y python-pip')
        else:
            raise NotImplementedError('Unknown pip bootstrap method: %s' % r.env.bootstrap_method)
            
        r.sudo('pip install --upgrade pip')
        r.sudo('pip install --upgrade virtualenv')

    @task
    def clean_virtualenv(self, virtualenv_dir=None):
        r = self.local_renderer
        with self.settings(warn_only=True):
            print('Deleting old virtual environment...')
            r.sudo('rm -Rf {virtualenv_dir}')
        
    @task
    def has_virtualenv(self):
        """
        Returns true if the virtualenv tool is installed.
        """
        with self.settings(warn_only=True):
            ret = self.run('which virtualenv').strip()
            return bool(ret)
    
    @task
    def virtualenv_exists(self, virtualenv_dir=None):
        """
        Returns true if the virtual environment has been created.
        """
        r = self.local_renderer
        ret = True
        with self.settings(warn_only=True):
            ret = r.run('ls {virtualenv_dir}') or ''
            ret = 'cannot access' not in ret.strip().lower()
            
        if self.verbose:
            if ret:
                print('Yes')
            else:
                print('No')
            
        return ret
    
    @task
    def init(self):
        """
        Creates the virtual environment.
        """
        r = self.local_renderer
        
#         if self.virtualenv_exists():
#             print('virtualenv exists')
#             return
        
        print('Creating new virtual environment...')
        with self.settings(warn_only=True):
            cmd = '[ ! -d {virtualenv_dir} ] && virtualenv --no-site-packages {virtualenv_dir} || true'
            if self.is_local:
                r.run(cmd)
            else:
                r.sudo(cmd)
    
    @task
    def set_permissions(self):
        r = self.local_renderer
        if not self.is_local and r.env.check_permissions:
            r.sudo('chown -R {pip_user}:{pip_group} {virtualenv_dir}')
            r.sudo('chmod -R {pip_perms} {virtualenv_dir}')
    
    def get_combined_requirements(self, requirements=None):
        """
        Returns all requirements files combined into one string.
        """
        
        requirements = requirements or self.env.requirements
        
        def iter_lines(fn):
            with open(fn, 'r') as fin:
                for line in fin.readlines():
                    line = line.strip()
                    if not line or line.startswith('#'):
                        continue
                    yield line
        
        content = []
        if isinstance(requirements, (tuple, list)):
            for f in requirements:
                f = self.find_template(f)
                content.extend(list(iter_lines(f)))
        else:
            assert isinstance(requirements, basestring)
            f = self.find_template(requirements)
            content.extend(list(iter_lines(f)))
        
        return '\n'.join(content)
    
    @task
    def update_install(self, **kwargs):
        r = self.local_renderer
        
        options = [
            'requirements',
            'virtualenv_dir',
            'user',
            'group',
            'perms',
        ]
        for option in options:
            setattr(r.env, option, kwargs.pop(option, None) or getattr(r.env, option))
        
        # Make sure pip is installed.
        self.bootstrap()
        
        # Make sure our virtualenv is installed.
        self.init()
        
        # Collect all requirements.
        tmp_fn = r.write_temp_file(self.get_combined_requirements(requirements=r.env.requirements))
        
        # Copy up our requirements.
        r.env.pip_remote_requirements_fn = '/tmp/pip-requirements.txt'
        r.put(
            local_path=tmp_fn,
            remote_path=r.env.pip_remote_requirements_fn,
        )
        
        # Ensure we're always using the latest pip.
        if self.is_local:
            r.run('{virtualenv_dir}/bin/pip install -U pip')
        else:
            r.sudo('{virtualenv_dir}/bin/pip install -U pip')
        
        cmd = "{virtualenv_dir}/bin/pip install -r {pip_remote_requirements_fn}"
        if self.is_local:
            r.run(cmd)
        else:
            r.sudo(cmd)
            
        if not self.is_local and r.env.check_permissions:
            self.set_permissions()

    @task
    def record_manifest(self):
        """
        Called after a deployment to record any data necessary to detect changes
        for a future deployment.
        """
        manifest = super(PIPSatchel, self).record_manifest()
        manifest['all-requirements'] = self.get_combined_requirements()
        if self.verbose:
            pprint(manifest, indent=4)
        return manifest
    
    @task(precursors=['packager', 'user'])
    def configure(self, *args, **kwargs):
        
        # Necessary to make warning message go away.
        # http://stackoverflow.com/q/27870003/247542
        self.genv['sudo_prefix'] += '-H '
        
        self.update_install(*args, **kwargs)
        
pip = PIPSatchel()



"""
Tests for the common module.
"""
from __future__ import print_function

import os
import sys
import tempfile
import unittest
import getpass
from pprint import pprint

# try:
#     import pytest
# except ImportError:
#     pass

from burlap import load_yaml_settings
from burlap.common import CMD_VAR_REGEX, CMD_ESCAPED_VAR_REGEX, shellquote, all_satchels, Satchel, env, get_satchel, clear_state
from burlap.decorators import task
from burlap.tests.base import TestCase

sys.path.insert(0, os.path.dirname(os.path.realpath(__file__)))

class _TestSatchel(Satchel):
    
    name = 'test'
    
    def configure(self):
        pass

class CommonTests(TestCase):

    def get_test_satchel(self):
        test = _TestSatchel()
        test.genv.hosts = ['localhost']
        test.genv.host_string = test.genv.hosts[0]
        return test

    def setUp(self):
        super(CommonTests, self).setUp()
                
        # Importing ourself register us in sys.modules, which burlap uses to track satchels.
        # This is necessary to instantiate this satchel when running this testcase separately.
        import test_common # pylint: disable=import-self

        env.hosts = ['localhost']
        env.host_string = env.hosts[0]
        env.user = getpass.getuser()
        env.always_use_pty = False
    
    def test_shellquote(self):
         
        s = """# /etc/cron.d/anacron: crontab entries for the anacron package
     
    SHELL=/bin/bash
    PATH=/usr/local/sbin:/usr/local/bin:/sbin:/bin:/usr/sbin:/usr/bin
     
    # minute hour day month weekday (0-6, 0 = Sunday) user command
     
    */5 * * * *   root    {command}
    """
     
        s = shellquote(s)
         
    def test_format_regex(self):
        
        test = self.get_test_satchel()

        assert CMD_VAR_REGEX.findall('{cmd} {host}') == ['cmd', 'host']
         
        assert CMD_VAR_REGEX.findall("{cmd} {host} | {awk_cmd} '{{ print $1 }}'") == ['cmd', 'host', 'awk_cmd']
     
        assert CMD_ESCAPED_VAR_REGEX.findall("{cmd}} {{ print hello }}") == ['{{ print hello }}']
     
        r = test.local_renderer
        r.env.host = 'myhost'
        r.local("getent {host} | awk '{{ print $1 }}'", dryrun=1)
         
        s = "rsync --recursive --verbose --perms --times --links --compress --copy-links {exclude_str}  ' \
            '--delete --delete-before --force {rsync_auth} {rsync_source_dir} {rsync_target_host}{rsync_target_dir}"
        assert CMD_VAR_REGEX.findall(s) == ['exclude_str', 'rsync_auth', 'rsync_source_dir', 'rsync_target_host', 'rsync_target_dir']
     
    def test_settings_include(self):
        try:
            os.makedirs('/tmp/burlap_test/roles/all')
        except OSError:
            pass
             
        try:
            os.makedirs('/tmp/burlap_test/roles/prod')
        except OSError:
            pass
             
        open('/tmp/burlap_test/roles/all/settings.yaml', 'w').write("""
only_all_param: "just in all"
overridden_by_prod: 123
overridden_by_local: slkdjflsk
    """)
        open('/tmp/burlap_test/roles/prod/settings.yaml', 'w').write("""inherits: all
overridden_by_prod: 'prod'
only_prod_param: 7891
overriden_by_include: 7892
overridden_by_local: oiuweoiruwo
#includes: [settings_include1.yaml]
includes: [settings_include2.yaml]
    """)
        open('/tmp/burlap_test/roles/prod/settings_include2.yaml', 'w').write("""
overriden_by_include: xyz
overridden_by_local: ovmxlkfsweirwio
    """)
        open('/tmp/burlap_test/roles/prod/settings_local.yaml', 'w').write("""
overridden_by_local: 'hello world'
includes: [settings_include3.yaml]
    """)
        open('/tmp/burlap_test/roles/prod/settings_include3.yaml', 'w').write("""
set_by_include3: 'some special setting'
    """)
        os.chdir('/tmp/burlap_test')
        config = load_yaml_settings(name='prod')
         
        assert config['includes'] == ['settings_include2.yaml', 'settings_include3.yaml']
        assert config['only_all_param'] == 'just in all'
        assert config['overridden_by_prod'] == 'prod'
        assert config['only_prod_param'] == 7891
        assert config['overriden_by_include'] == 'xyz'
        assert config['overridden_by_local'] == 'hello world'
        assert config['set_by_include3'] == 'some special setting'
     
    def test_renderer(self):
        
        test = self.get_test_satchel()

        # Confirm renderer is cached.
        r1 = test.local_renderer
        r2 = test.local_renderer
        assert r1 is r2
         
        # Confirm clear method.
        test.clear_local_renderer()
        r3 = test.local_renderer
        assert r1 is not r3
         
        r = r3
        env.clear()
        assert r.genv is env
         
        # Confirm local env var gets renderered.
        r.env.var1 = 'a'
        assert r.format('{var1}') == 'a'
         
        # Confirm global env var in local namespace gets rendered.
        env.test_var2 = 'b'
        assert r.format('{var2}') == 'b'
         
        # Confirm global env var in global namespace gets rendered.
        env.test_var2 = 'b2'
        assert r.format('{test_var2}') == 'b2'
         
        # Confirm global env var overridden in local namespace get rendered.
        env.apache_var3 = '0'
        r.env.apache_var3 = 'c'
        assert r.format('{apache_var3}') == 'c'
         
        # Confirm recursive template variables get rendered.
        r.env.some_template = '{target_value}'
        r.env.target_value = 'd'
        assert r.format('{some_template}') == 'd'
         
        class ApacheSatchel(Satchel):
             
            name = 'apache'
             
            def configure(self):
                pass
             
        apache = ApacheSatchel()
        r = apache.local_renderer
        r.env.application_name = 'someappname'
        r.env.site = 'sitename'
        r.env.wsgi_path = '/usr/local/{apache_application_name}/src/wsgi/{apache_site}.wsgi'
        assert r.format(r.env.wsgi_path) == '/usr/local/someappname/src/wsgi/sitename.wsgi'

    def test_iter_sites(self):
        
        test = self.get_test_satchel()
         
        env.sites = {
            'site1': {'apache_ssl': False},
            'site2': {'apache_ssl': True},
        }
         
        lst = list(test.iter_sites())
        print('lst:', lst)
        assert len(lst) == 2
         
        lst = list(test.iter_sites(site='site2'))
        print('lst:', lst)
        assert len(lst) == 1
     
    def test_append(self):

        test = self.get_test_satchel()

        test.genv.host_string = 'localhost'

        _, fn = tempfile.mkstemp()
        text = '[{rabbit, [{loopback_users, []}]}].'
        test.append(filename=fn, text=text)
        content = open(fn).read()
        print('content0:', content)
        assert content.count(text) == 1

        # Confirm duplicate lines are appended.
        test.append(filename=fn, text=text)
        content = open(fn).read()
        print('content1:', content)
        assert content.count(text) == 1

    def test_set_verbose(self):
        from burlap.common import set_verbose, get_verbose
        
        set_verbose(True)
        assert get_verbose()
        
        set_verbose(False)
        assert not get_verbose()
        
        set_verbose(1)
        assert get_verbose()
        
        set_verbose(0)
        assert not get_verbose()

    def test_satchel_ordering(self):
        from burlap.deploy import preview, init_plan_data_dir
        
        # Purge any pre-existing satchels from global registeries so we only get results for our custom satchels.
        clear_state()

        # These test satchels should be dependent in the order c<-a<-b
         
        class ASatchel(Satchel):
            name = 'a'
            def set_defaults(self):
                self.env.param = 123
            @task(precursors=['c'])
            def configure(self):
                pass
         
        class BSatchel(Satchel):
            name = 'b'
            def set_defaults(self):
                self.env.param = 123
            @task(precursors=['a', 'c'])
            def configure(self):
                pass
         
        class CSatchel(Satchel):
            name = 'c'
            def set_defaults(self):
                self.env.param = 123
            @task
            def configure(self):
                pass
         
        a_satchel = ASatchel()
        b_satchel = BSatchel()
        c_satchel = CSatchel()
        try:
     
            assert set(all_satchels) == set(['A', 'B', 'C'])
             
            assert init_plan_data_dir() == '.burlap/plans'
             
            env.ROLE = 'local'
            components, plan_funcs = preview(components=['A', 'B', 'C'], enable_plans=False, force=True)
            expected_components = ['C', 'A', 'B']
            print()
            print('components:', components)
            print('expected_components:', expected_components)
            print('plan_funcs:', plan_funcs)
            task_names = [_0 for _0, _1 in plan_funcs]
            assert components == expected_components
            assert task_names == ['c.configure', 'a.configure', 'b.configure']
         
        finally:
            a_satchel.unregister()
            del a_satchel
            with self.assertRaises(KeyError):
                get_satchel('a')
#             import gc
#             refs = gc.get_referrers(a_satchel)
#             print('refs:', refs)
            b_satchel.unregister()
            c_satchel.unregister()

    def test_state_clearing(self):
        from burlap.common import get_state, clear_state, set_state, all_satchels
        
        actual = sorted(all_satchels.keys())
        print('all_satchels.a:', actual)
        expected = [
            'APACHE', 'AVAHI', 'BLUETOOTH', 'BUILDBOT', 'CELERY', 'CLOUDFRONT', 'CRON',
            'DEBUG', 'DEPLOYMENTNOTIFIER', 'DJ', 'EC2MONITOR', 'ELASTICSEARCH', 'FILE',
            'GIT', 'GITCHECKER', 'GITTRACKER', 'GPSD', 'GROUP', 'HOST', 'HOSTNAME', 'HOSTSFILE',
            'IP', 'JIRAHELPER', 'JSHINT', 'LOGINNOTIFIER', 'MOTION', 'MYSQL', 'MYSQLCLIENT', 'NM',
            'NTPCLIENT', 'PACKAGER', 'PHANTOMJS', 'PIP', 'POSTFIX', 'POSTGRESQL', 'POSTGRESQLCLIENT', 'PROJECT',
            'RABBITMQ', 'RPI', 'RSYNC', 'S3', 'SELENIUM', 'SERVICE', 'SNORT', 'SOFTWARERAID',
            'SSHNICE', 'SSL', 'SUPERVISOR', 'TARBALL', 'TIMEZONE', 'UBUNTUMULTIVERSE',
            'UNATTENDEDUPGRADES', 'USER', 'VAGRANT', 'VIRTUALBOX',
        ]
        assert actual == expected
        
        burlap_state = get_state()
        print('burlap_state:')
        pprint(burlap_state, indent=4)
        
        clear_state()
        print('all_satchels.b:', sorted(all_satchels.keys()))
        assert len(all_satchels) == 0
        
        set_state(burlap_state)
        print('all_satchels.c:', sorted(all_satchels.keys()))
        assert len(all_satchels) == len(expected)
        
    def test_runs_once_clear(self):
        from fabric.api import runs_once
        from burlap.debug import debug
        from burlap.common import runs_once_methods
        
        a = ['abc']
        
        @runs_once
        def test_func():
            return a[0]
        
        print('a')
        assert test_func() == 'abc'
        
        a[0] = 'xyz'
        del test_func.return_value
        
        print('b')
        assert test_func() == 'xyz'
        
        print('c')
        a[0] = 'hhh'
        assert test_func() == 'xyz'
        
        print('debug.shell:', debug.shell)
        #assert hasattr(debug.shell, 'wrapped')
        print('runs_once_methods:', runs_once_methods)
        

if __name__ == '__main__':
    unittest.main()



"""
Wrapper around the Motion service.

http://www.lavrsen.dk/foswiki/bin/view/Motion/WebHome
"""
from __future__ import print_function

from burlap import ServiceSatchel
from burlap.constants import *
from burlap.decorators import task

class MotionSatchel(ServiceSatchel):
    
    name = 'motion'
    
    @property
    def packager_system_packages(self):
        return {
            FEDORA: ['motion'],
            UBUNTU: ['motion'],
        }
    
    def set_defaults(self):
        self.env.notify_enabled = False
        self.env.service_commands = {
            START:{
                FEDORA: 'systemctl start motion.service',
                UBUNTU: 'service motion start',
            },
            STOP:{
                FEDORA: 'systemctl stop motion.service',
                UBUNTU: 'service motion stop',
            },
            DISABLE:{
                FEDORA: 'systemctl disable motion.service',
                UBUNTU: 'chkconfig motion off',
            },
            ENABLE:{
                FEDORA: 'systemctl enable motion.service',
                UBUNTU: 'chkconfig motion on',
            },
            RESTART:{
                FEDORA: 'systemctl restart motion.service',
                UBUNTU: 'service motion restart; sleep 5',
            },
            STATUS:{
                FEDORA: 'systemctl status motion.service',
                UBUNTU: 'service motion status',
            },
        }    
    
    @task(precursors=['packager', 'user'])
    def configure(self):
        pass
    
motion = MotionSatchel()



from __future__ import print_function

from burlap.constants import *
from burlap.mysql import MySQLSatchel, MYSQLD_SAFE
from burlap.tests.functional_tests.base import TestCase
from burlap.system import distrib_id, distrib_release

MYSQL_ROOT_PASSWORD = 's3cr3t'

class MySQLTests(TestCase):
        
    def test_set_root_password_mysqld_safe(self):
        mysql = MySQLSatchel()
        try:
            mysql.verbose = True
            
            # https://docs.travis-ci.com/user/ci-environment/
            # Travis-CI uses Ubuntu 14.04, which has some odd default MySQL packages that we need to remove
            # before we can install the correct ones.
            # http://askubuntu.com/a/489817/13217
    #         os_ver = mysql.os_version
    #         if os_ver.distro == UBUNTU and os_ver.release == '14.04':
    #             mysql.sudo('DEBIAN_FRONTEND=noninteractive apt-get -yq purge mysql*')
    #             mysql.sudo('DEBIAN_FRONTEND=noninteractive apt-get -yq autoremove')
    #             mysql.sudo('DEBIAN_FRONTEND=noninteractive apt-get -yq autoclean')
    #             mysql.sudo('DEBIAN_FRONTEND=noninteractive apt-get -yq update')
    #             mysql.sudo('DEBIAN_FRONTEND=noninteractive apt-get -yq install mysql-client-core-5.5')
                
            mysql.install_packages()
            mysql.env.root_username = 'root'
            mysql.env.root_password = mysql.env.db_root_password = MYSQL_ROOT_PASSWORD
            mysql.set_root_password(method=MYSQLD_SAFE)
            ret = mysql.execute('SHOW DATABASES;', use_sudo=True)
            print('ret:', ret)
        finally:
            mysql.purge_packages()

    #TODO:reproduce "Row size too large" error and fix by enabling Barracuda
    def _test_barracuda(self):
        d_id = distrib_id()
        print('d_id:', d_id)
        d_release = distrib_release()
        print('d_release:', d_release)
        mysql = MySQLSatchel()
        mysql.install_packages()
        try:
            mysql_version = mysql.get_mysql_version()
            mysql_conf_path = mysql.conf_path
            print('mysql_version:', mysql_version)
            if d_id == UBUNTU:
                if d_release == '14.04':
                    assert mysql_version == '5.6'
                    assert mysql_conf_path == '/etc/mysql/my.cnf'
                elif d_release == '16.04':
                    assert mysql_version == '5.7'
                    assert mysql_conf_path == '/etc/mysql/mysql.conf.d/mysqld.cnf'

            # Reproduce "Row size too large" error.
            # https://bugs.mysql.com/bug.php?id=69336
            mysql.write_to_file(
                content='''
#!/bin/bash

n=$1
: ${n:=228}

cat<<GO > m2b.sql
CREATE DATABASE IF NOT EXISTS test;
use test;
set global innodb_file_format=Barracuda;
set global innodb_file_format_max=Barracuda;
set global innodb_file_per_table=1;

drop table if exists foo;
create table foo (
id int auto_increment primary key,
v varchar(32)
GO

for i in `seq 1 $n` ; do
   echo ", col$i text"
done >> m2b.sql
echo ") ENGINE=MyISAM;" >> m2b.sql

echo "alter table foo engine=innodb row_format=dynamic;" >> m2b.sql
''',
                fn='/tmp/create.sh')
            mysql.run('cd /tmp; bash /tmp/create.sh')
            mysql.sudo('cd /tmp; mysql < m2b.sql')

            
        finally:
            #mysql.purge_packages()
            pass
